from __future__ import annotations

r"""A set of generic classes and corresponding functions to iterate and index them.


The main goal of having this module is to allow for flexible modeling of collections with *most* python builtin
collections (e.g. `list`, `tuple`, `generator`) or even `Callable`\ s without awkward `isinstance` calls in
other modules of the library.

"""
from typing import Any, Callable, Dict, Iterable, Mapping, Sequence, Union

__all__ = [
    "IterableMapping",  # A mapping combining dicts, lists, tuples, and generators
    "GenericMapping",
    # An iterable mapping or a callable (e.g. function/functor)
    # GenericMapping functions (apply to GMapping and IterableMapping)
    "gmap",  # An index on an IterableMapping or a call on a Callable
    "gget",
    # Like `gmap` but can return a default value
    # IterableMapping functions (apply only to IterableMappings)
    "ienumerate",  # enumerates a mapping as key-value pairs
    "iitems",  # enumerates a mapping as key-value pairs
    "ivalues",  # enumerates all values in a mapping (no keys)
    "ikeys",  # enumerates all keys in a mapping (no values)
    "iget",  # Similar to `gget` but applies only to IterableMappings
]
# Generic classes that work with a variety of python mappings like collections or callables
IterableMapping = Union[Mapping, Sequence]
"""Something that can be iterated upon with Key-value pairs (e.g. list, dict, tuple)."""
GenericMapping = Union[Callable[[Any], Any], IterableMapping]
"""Something that can be indexed using [] or called using ()"""


def gmap(group: GenericMapping, param: Any) -> Any:
    """Calls or indexes the group by the param

    Args:
        group: Either a Callable or a Mapping
        param: The parameters to use for mapping

    Examples:

        >>> gmap([1, 23, 44], 1)
        23
        >>> gmap({'a': 3, 'b': 5, 'c': 4}, 'c')
        4
        >>> gmap(lambda x: 3*x, 20)
        60

    Returns:

    """

    if hasattr(group, "__call__"):
        return group(param)  # type: ignore

    return group[param]  # type: ignore


def iget(x: IterableMapping, _key: Any, default=None) -> Any:
    """Get an item from an IterableMapping

    Args:
        x: the generic mapping
        _key: key (must be immutable)
        default: default value if no value attached with the key is found

    Examples:

        Example with a list

        >>> [iget([10, 20, 30], _) for _ in (0, 2,1, -1, 4)]
        [10, 30, 20, 30, None]

        Example with a dictionary

        >>> [iget({'a':10, 'b':20, 'c':30}, _) for _ in ('a', 'c','b', -1, 'd')]
        [10, 30, 20, None, None]

        Example with a tuple

        >>> [iget((10, 20, 30), _) for _ in (0, 2,1, -1, 4)]
        [10, 30, 20, 30, None]

        Example with a generator

        >>> [iget(range(10, 40, 10), _) for _ in (0, 2,1, -1, 4)]
        [10, 30, 20, 30, None]

    Returns:

    """

    if isinstance(x, Dict):
        return x.get(_key, default)

    try:
        return x[_key]  # type: ignore
    except IndexError:
        return default


def gget(x: GenericMapping, _key: Any, default=None) -> Any:
    """Get an item from an IterableMapping

    Args:
        x: the generic mapping
        _key: key (must be immutable)
        default: default value if no value attached with the key is found

    Examples:

        Example with a list

        >>> [gget([10, 20, 30], _) for _ in (0, 2,1, -1, 4)]
        [10, 30, 20, 30, None]

        Example with a dictionary

        >>> [gget({'a':10, 'b':20, 'c':30}, _) for _ in ('a', 'c','b', -1, 'd')]
        [10, 30, 20, None, None]

        Example with a tuple

        >>> [gget((10, 20, 30), _) for _ in (0, 2,1, -1, 4)]
        [10, 30, 20, 30, None]

        Example with a generator

        >>> [gget(range(10, 40, 10), _) for _ in (0, 2,1, -1, 4)]
        [10, 30, 20, 30, None]

    Returns:

    """

    if hasattr(x, "apply"):
        # noinspection PyBroadException
        try:
            return x(_key)  # type: ignore

        except Exception:
            return default

    else:
        return iget(x, _key, default)  # type: ignore


def ienumerate(x: IterableMapping) -> Iterable[tuple[Any, Any]]:
    """Enumerates a GenericMapping.

    Args:
        x (IterableMapping): A generic mapping (see `GenericMapping`)

    Examples:

        Example with a list

        >>> for k, cutoff_utility in ienumerate([10, 20, 30]): print(k, cutoff_utility, end='-')
        0 10-1 20-2 30-

        Example with a dictionary

        >>> for k, cutoff_utility in ienumerate({'a': 10, 'b': 20, 'c': 30}): print(k, cutoff_utility, end='-')
        a 10-b 20-c 30-

        Example with a tuple

        >>> for k, cutoff_utility in ienumerate((10, 20, 30)): print(k, cutoff_utility, end='-')
        0 10-1 20-2 30-

        Example with a generator

        >>> for k, cutoff_utility in ienumerate(range(10, 40, 10)): print(k, cutoff_utility, end='-')
        0 10-1 20-2 30-

    Returns:
        a generator/iterator with tuples of key-value pairs.

    """

    if isinstance(x, Dict):
        return x.items()

    else:
        return enumerate(x)


iitems = ienumerate


def ivalues(x: IterableMapping) -> Iterable[Any]:
    """Returns all keys of the iterable.

    Args:
        x (IterableMapping): A generic mapping (see `GenericMapping`)

    Examples:

        Example with a list

        >>> for k in ivalues([10, 20, 30]): print(k, end='-')
        10-20-30-

        Example with a dictionary

        >>> for k in ivalues({'a': 10, 'b': 20, 'c': 30}): print(k, end='-')
        10-20-30-

        Example with a tuple

        >>> for k in ivalues((10, 20, 30)): print(k, end='-')
        10-20-30-

        Example with a generator

        >>> for k in ivalues(range(10, 40, 10)): print(k, end='-')
        10-20-30-

    Returns:
        a generator/iterator with tuples of key-value pairs.

    """

    if isinstance(x, Dict):
        return list(x.values())

    else:
        return x


def ikeys(x: IterableMapping) -> Iterable[Any]:
    """Returns all keys of the iterable.

    Args:
        x (IterableMapping): A generic mapping (see `GenericMapping`)

    Examples:

        Example with a list

        >>> for k in ikeys([10, 20, 30]): print(k, end='-')
        0-1-2-

        Example with a dictionary

        >>> for k in ikeys({'a': 10, 'b': 20, 'c': 30}): print(k, end='-')
        a-b-c-

        Example with a tuple

        >>> for k in ikeys((10, 20, 30)): print(k, end='-')
        0-1-2-

        Example with a generator

        >>> for k in ikeys(range(10, 40, 10)): print(k, end='-')
        0-1-2-

    Returns:
        a generator/iterator with tuples of key-value pairs.

    """

    if isinstance(x, Dict):
        return list(x.keys())
    else:
        return range(len(x))  # type: ignore



from __future__ import annotations

from ..components.acceptance import (
    AcceptAnyRational,
    AcceptBetterRational,
    AcceptNotWorseRational,
)
from ..components.offering import CABOfferingPolicy
from .modular.mapneg import MAPNegotiator

__all__ = [
    "CABNegotiator",
    "CARNegotiator",
    "CANNegotiator",
]


class CANNegotiator(MAPNegotiator):
    """
    Conceding Accepting Not Worse Strategy (optimal, complete, but not an equilibirum)

    Args:
         name: Negotiator name
         parent: Parent controller if any
         preferences: The preferences of the negotiator
         ufun: The ufun of the negotiator (overrides prefrences)
         owner: The `Agent` that owns the negotiator.

    """

    def __init__(self, *args, **kwargs):
        kwargs["acceptance"] = AcceptNotWorseRational()
        kwargs["offering"] = CABOfferingPolicy()
        super().__init__(*args, **kwargs)


class CABNegotiator(MAPNegotiator):
    """
    Conceding Accepting Better Strategy (optimal, complete, but not an equilibirum)

    Args:
         name: Negotiator name
         parent: Parent controller if any
         preferences: The preferences of the negotiator
         ufun: The ufun of the negotiator (overrides prefrences)
         owner: The `Agent` that owns the negotiator.

    """

    def __init__(self, *args, **kwargs):
        kwargs["acceptance"] = AcceptBetterRational()
        kwargs["offering"] = CABOfferingPolicy()
        super().__init__(*args, **kwargs)


class CARNegotiator(MAPNegotiator):
    """
    Conceding Accepting Rational Strategy (neither complete nor an equilibrium)

    Args:
         name: Negotiator name
         parent: Parent controller if any
         preferences: The preferences of the negotiator
         ufun: The ufun of the negotiator (overrides prefrences)
         owner: The `Agent` that owns the negotiator.

    """

    def __init__(self, *args, **kwargs):
        kwargs["acceptance"] = AcceptAnyRational()
        kwargs["offering"] = CABOfferingPolicy()
        super().__init__(*args, **kwargs)



#!/usr/bin/env python
"""
Extra helpers
"""
from __future__ import annotations

import itertools
import socket
from typing import Any, Callable, Iterable, Sequence

from ..protocols import HasMinMax

__all__ = [
    "monotonic_minmax",
    "monotonic_multi_minmax",
    "nonmonotonic_multi_minmax",
    "nonmonotonic_minmax",
    "make_callable",
    "get_free_tcp_port",
]


def nonmonotonic_minmax(
    input: Iterable, f: Callable[[Any], float]
) -> tuple[float, float]:
    """Finds the limits of a function `f` for the input assuming that it is non-monotonic and input is iterable"""
    mn, mx = float("inf"), float("-inf")
    fmn, fmx = float("-inf"), float("inf")
    for x in input:
        fx = float(f(x))
        if fx < mn:
            mn, fmn = x, fx
        if fx > mx:
            mx, fmx = x, fx
    return fmn, fmx


def nonmonotonic_multi_minmax(
    input: Iterable[Iterable], f: Callable[[Any], float]
) -> tuple[float, float]:
    """Finds the limits of a function `f` for each input assuming that it is non-monotonic and input is iterable"""
    return nonmonotonic_minmax(itertools.product(*input), f)


def monotonic_minmax(
    input: HasMinMax, f: Callable[[Any], float]
) -> tuple[float, float]:
    """Finds the limits of a function `f` for the input assuming that it is monotonic and input has `min_value` and `max_value` members"""
    a, b = input.min_value, input.max_value
    fa, fb = f(a), f(b)
    if fb < fa:
        return float(fb), float(fa)
    return float(fa), float(fb)


def monotonic_multi_minmax(
    input: Iterable[HasMinMax], f: Callable[[Any], float]
) -> tuple[float, float]:
    """Finds the limits of a function `f` for the input assuming that it is monotonic and each input has `min_value` and `max_value` members"""
    vals = [(i.min_value, i.max_value) for i in input]
    return nonmonotonic_minmax(itertools.product(*vals), f)


def remove_qoutes(s: str) -> str:
    return s.replace('"', "~`")


def recover_qoutes(s: str) -> str:
    return s.replace("~`", '"')


def make_callable(x: dict | Sequence | Callable | None) -> Callable:
    """
    Converts its input to a callable (i.e. can be called using () operator).


    Examples:

        >>> make_callable(lambda  x: 2 * x) (4)
        8

        >>> make_callable(dict(a=1, b=3))("a")
        1

        >>> make_callable((3, 4, 5))(1)
        4
    """
    if x is None:
        return lambda a: a
    if isinstance(x, Callable):
        return x
    return lambda a: x[a]


def get_free_tcp_port():
    tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    tcp.bind(("", 0))
    _, port = tcp.getsockname()
    tcp.close()
    return port



from __future__ import annotations

import sys

from attrs import define

from ..common import ThreadState
from .base import LocalOfferingConstraint

__all__ = ["RepeatFinalOfferOnly"]


@define
class RepeatFinalOfferOnly(LocalOfferingConstraint):
    n: int = sys.maxsize

    def __call__(
        self,
        state: ThreadState,
        history: list[ThreadState],
    ) -> bool:
        offer = state.new_offer
        if not offer:
            return False
        outcomes = [_.new_offer for _ in history]
        if outcomes:
            for a, b in zip(outcomes[:-1], outcomes[1:]):
                if a == offer:
                    break
                if a == b:
                    return False
        past = set(outcomes)
        if history and len(past) >= self.n:
            past = past.difference({history[-1].new_offer})
        return offer not in past



from __future__ import annotations

import copy
import itertools
import json
import numbers
import sys
import xml.etree.ElementTree as ET
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Callable, Iterable, Sequence

from negmas.outcomes.contiguous_issue import ContiguousIssue
from negmas.warnings import warn_if_slow

if TYPE_CHECKING:
    from .common import Outcome

import numpy as np

from negmas.helpers import PathLike

from .base_issue import DiscreteIssue, Issue, make_issue
from .outcome_ops import outcome2dict

__all__ = [
    "generate_issues",
    "issues_from_genius",
    "issues_from_geniusweb",
    "issues_from_xml_str",
    "issues_from_geniusweb_json_str",
    "issues_to_genius",
    "issues_to_xml_str",
    "issues_from_outcomes",
    "num_outcomes",
    "enumerate_issues",
    "enumerate_discrete_issues",
    "discretize_and_enumerate_issues",
    "sample_issues",
    "sample_outcomes",
    "combine_issues",
]


DUMMY_ISSUE_NAME = "DUMMY_ISSUE"


def num_outcomes(issues: Sequence[Issue]) -> int | float:
    """
    Returns the total number of outcomes in a set of issues.
    """

    n = 1

    for issue in issues:
        n *= issue.cardinality

    return n


def enumerate_discrete_issues(issues: Sequence[DiscreteIssue]) -> list[Outcome]:
    """
    Enumerates all outcomes of this set of discrete issues if possible.

    Args:
        issues: A list of issues

    Returns:
        list of outcomes
    """
    n = num_outcomes(issues)
    warn_if_slow(n, "Enumerating large OS (discrete)")
    return list(itertools.product(*(_.all for _ in issues)))


def sample_outcomes(
    issues: Sequence[Issue],
    n_outcomes: int | None = None,
    min_per_dim: int = 5,
    expansion_policy=None,
) -> list[Outcome]:
    """
    Discretizes the issue space and returns either a predefined number of outcomes or uniform samples.

    Args:
        issues: The issues describing the issue space to be discretized
        n_outcomes: If None then exactly `min_per_dim` bins will be used for every continuous dimension and all outcomes
        will be returned
        min_per_dim: Max levels of discretization per dimension
        expansion_policy: None or 'repeat' or 'null' or 'no'. If repeat, then some of the outcomes will be repeated
        if None or 'no' then no expansion will happen if the total number of outcomes is less than
        n_outcomes: If 'null' then expansion will be with None values

    Returns:
        list of outcomes

    Examples:

        enumberate the whole space

        >>> from negmas.outcomes import make_issue
        >>> issues = [make_issue(values=(0.0, 1.0), name='Price'), make_issue(values=['a', 'b'], name='Name')]
        >>> sample_outcomes(issues=issues)
        [(0.0, 'a'), (0.0, 'b'), (0.25, 'a'), (0.25, 'b'), (0.5, 'a'), (0.5, 'b'), (0.75, 'a'), (0.75, 'b'), (1.0, 'a'), (1.0, 'b')]

        enumerate with sampling for very large space (we have 10 outcomes in the discretized space)

        >>> from negmas.outcomes import make_issue
        >>> issues = [make_issue(values=(0.0, 1.0), name='Price'), make_issue(values=['a', 'b'], name='Name')]
        >>> issues[0].is_continuous()
        True
        >>> sampled=sample_outcomes(issues=issues, n_outcomes=5)
        >>> len(sampled)
        5
        >>> len(set(sampled))
        5

        >>> from negmas.outcomes import make_issue
        >>> issues = [make_issue(values=(0, 1), name='Price'), make_issue(values=['a', 'b'], name='Name')]
        >>> issues[0].is_continuous()
        False
        >>> sampled=sample_outcomes(issues=issues, n_outcomes=5)
        >>> len(sampled)
        4
        >>> len(set(sampled))
        4

    """
    from negmas.outcomes import enumerate_discrete_issues

    issues = list(issues)
    issues = [copy.deepcopy(_) for _ in issues]
    continuous = []
    indx = []
    discrete = []
    n_disc = 0

    for i, issue in enumerate(issues):
        if issue.is_continuous():
            continuous.append(issue)
            indx.append(i)
        else:
            discrete.append(issue)
            n_disc += issue.cardinality

    if len(continuous) > 0:
        if n_outcomes is not None:
            n_per_issue = max(min_per_dim, int(n_outcomes - n_disc) // len(continuous))
        else:
            n_per_issue = min_per_dim

        for i, issue in enumerate(continuous):
            issues[indx[i]] = make_issue(
                name=issue.name,
                values=list(
                    np.linspace(
                        issue.min_value, issue.max_value, num=n_per_issue, endpoint=True
                    ).tolist()
                ),
            )

    cardinality = 1

    for issue in issues:
        cardinality *= issue.cardinality

    if cardinality == n_outcomes or n_outcomes is None:
        return list(
            enumerate_discrete_issues(issues)  # type: ignore I am  sure that the issues are all discrete by this point
        )

    if cardinality < n_outcomes:
        cardinality = int(cardinality)
        outcomes = list(
            enumerate_discrete_issues(issues)  # type: ignore I am  sure that the issues are all discrete by this point
        )

        if expansion_policy == "no" or expansion_policy is None:
            return outcomes
        elif expansion_policy == "null":
            return outcomes + [None] * (n_outcomes - cardinality)
        elif expansion_policy == "repeat":
            n_reps = n_outcomes // cardinality
            n_rem = n_outcomes % cardinality

            if n_reps > 1:
                for _ in range(n_reps):
                    outcomes += outcomes

            if n_rem > 0:
                outcomes += outcomes[:n_rem]

            return outcomes
    n_per_issue = 1 + int(n_outcomes ** (1 / len(issues)))
    vals = []
    n_found = 1
    for issue in issues:
        if n_per_issue < 2:
            vals.append([issue.rand()])
            continue
        if issue.cardinality < n_per_issue:
            vals.append(issue.all)
            n_found *= issue.cardinality
        else:
            vals.append(issue.rand_outcomes(n=n_per_issue, with_replacement=False))
            n_found *= n_per_issue
        if n_found >= n_outcomes:
            n_per_issue = 1
    outcomes = itertools.product(*vals)
    return list(outcomes)[:n_outcomes]


def _sample_issues(
    issues: Sequence[Issue],
    n: int,
    with_replacement,
    n_total,
    old_values,
    trial,
    max_trials,
) -> Iterable[Outcome]:
    if trial > max_trials:
        return old_values

    remaining = n - len(old_values)
    if remaining < 0:
        return set(list(old_values)[:n])
    if remaining == 0:
        return old_values

    samples = []

    for issue in issues:
        samples.append(
            issue.rand_outcomes(
                n=remaining, with_replacement=True, fail_if_not_enough=True
            )
        )

    _v = []

    for i in range(remaining):
        _v.append([s[i] for s in samples])

    new_values = []
    for value in _v:
        new_values.append(tuple(value))
    new_values = set(new_values)
    values = old_values.union(new_values)
    remaining = n - len(values)
    if remaining < 0:
        return set(list(values)[:n])
    if remaining < 1:
        return values
    return values.union(
        _sample_issues(
            issues,
            remaining,
            with_replacement,
            n_total,
            values,
            trial + 1,
            max_trials,
        )
    )


def sample_issues(
    issues: Sequence[Issue],
    n_outcomes: int,
    with_replacement: bool = True,
    fail_if_not_enough=True,
) -> Iterable[Outcome]:
    """
    Samples some outcomes from the outcome space defined by the list of issues.

    Args:

        issues: list of issues to sample from
        n_outcomes: The number of outcomes required
        with_replacement: Whether sampling is with replacement (allowing repetition)
        fail_if_not_enough: IF given then an exception is raised if not enough outcomes are available

    Returns:

        a list of outcomes

    Examples:

        >>> from negmas.outcomes import make_issue
        >>> issues = [make_issue(name='price', values=(0.0, 3.0)), make_issue(name='quantity', values=10)]

        Sampling outcomes as tuples

        >>> samples = sample_issues(issues=issues, n_outcomes=10)
        >>> len(samples) == 10
        True
        >>> type(samples[0]) == tuple
        True


    """
    n_total = num_outcomes(issues)

    if (
        n_total is not None
        and n_total != float("inf")
        and n_outcomes is not None
        and n_total < n_outcomes
        and fail_if_not_enough
        and not with_replacement
    ):
        raise ValueError(
            f"Cannot sample {n_outcomes} from a total of possible {n_total} outcomes"
        )

    if n_total is not None and n_total != float("inf") and n_outcomes is None:
        return enumerate_discrete_issues(issues=issues)  # type: ignore I know that these issues are discrete
    if n_total is None and n_outcomes is None:
        raise ValueError(
            f"Cannot sample unknown number of outcomes from continuous outcome spaces"
        )
    return list(
        _sample_issues(issues, n_outcomes, with_replacement, n_total, set(), 0, 10)
    )


def enumerate_issues(
    issues: Sequence[Issue],
    max_cardinality: int | float | None = None,
) -> list[Outcome]:
    """
    Enumerates the outcomes of a list of issues.

    Args:
        issues: The list of issues.
        max_cardinality: The maximum number of outcomes to return
    Returns:
        list of outcomes of the given type.
    """
    n = num_outcomes(issues)
    if isinstance(max_cardinality, float) and max_cardinality == float("inf"):
        max_cardinality = None

    if n is None and max_cardinality is not None:
        warn_if_slow(max_cardinality, "Enumerating a large OS")
    if max_cardinality is None:
        warn_if_slow(n, "Enumerating a continuous OS")

    if n is None:
        if max_cardinality is None:
            raise ValueError(
                "Cannot enumerate continuous issues without specifying `max_cardinality`"
            )
        return list(
            sample_issues(
                issues=issues,
                n_outcomes=int(max_cardinality),
                fail_if_not_enough=False,
                with_replacement=False,
            )
        )

    if max_cardinality is not None and n > max_cardinality:
        return sample_outcomes(issues=issues, n_outcomes=int(max_cardinality))
    else:
        return list(tuple(_) for _ in itertools.product(*(_.all for _ in issues)))


def issues_from_outcomes(
    outcomes: Sequence[Outcome] | int,
    numeric_as_ranges: bool = True,
    issue_names: list[str] | None = None,
) -> tuple[DiscreteIssue]:
    """
    Create a set of issues given some outcomes.

    Args:

        outcomes: A list of outcomes or the number of outcomes
        issue_names: If given, will be used as issue names, otherwise random issue names will be used
        numeric_as_ranges: If True, all numeric issues generated will have ranges that are defined by the minimum
                           and maximum values of that issue in the given outcomes instead of a list of the values
                           that appeared in them.

    Returns:

        a list of issues that include the given outcomes.

    Remarks:

        - The outcome space spanned by the generated issues can in principle contain many more possible outcomes
          than the ones given

    """

    if isinstance(outcomes, int):
        outcomes = [(_,) for _ in range(outcomes)]

    def convert_type(v, old, values):
        if isinstance(v, numbers.Integral) and not isinstance(old, numbers.Integral):
            return float(v)

        if not isinstance(v, numbers.Integral) and isinstance(old, numbers.Integral):
            for i, _ in enumerate(values):
                values[i] = float(_)

            return v

        if isinstance(v, str) and (isinstance(old, numbers.Number)):
            raise ValueError("a string after a number")

        if isinstance(old, str) and (isinstance(v, numbers.Number)):
            raise ValueError("a number after a string")

        return v

    names = None
    n_issues = None
    values = defaultdict(list)

    for i, o in enumerate(outcomes):
        if o is None:
            continue

        if issue_names is None:
            if isinstance(o, dict):
                issue_names = list(o.keys())
            else:
                issue_names = [f"i{_}" for _ in range(len(o))]

        if n_issues is not None and len(o) != n_issues:
            raise ValueError(
                f"Outcome {o} at {i} has {len(o)} issues but an earlier outcome had {n_issues} issues"
            )

        n_issues = len(o)
        if len(issue_names) != n_issues:
            raise ValueError(
                f"Outcome {i} ({o}) has {len(o)} values but we have {len(issue_names)} issue names"
            )

        o_dict = outcome2dict(o, issue_names)

        if names is not None and not all(a == b for a, b in zip(names, o_dict.keys())):
            raise ValueError(
                f"Outcome {o} at {i} has issues {list(o_dict.keys())} but an earlier outcome had issues {names}"
            )

        for i, v in enumerate(o):
            k = issue_names[i]
            if len(values[k]) > 0:
                try:
                    v = convert_type(v, values[k][-1], values[k])
                except ValueError as e:
                    raise ValueError(
                        f"Outcome {o} at {i} has value {v} for issue {k} which is incompatible with an earlier "
                        f"value {values[k][-1]} ({str(e)})"
                    )
            values[k].append(v)

    for k, vals in values.items():
        values[k] = sorted(list(set(vals)))

    if numeric_as_ranges:
        return tuple(  # type: ignore (seems  ok but not sure)
            make_issue(values=(v[0], v[-1]), name=n)
            if len(v) > 0
            and (isinstance(v[0], int))
            and all(a == b + 1 for a, b in zip(v[1:], v[:-1]))
            else make_issue(values=v, name=n)
            for n, v in values.items()
        )
    else:
        return tuple(make_issue(values=v, name=n) for n, v in values.items())  # type: ignore (seems  ok but not sure)


def issues_to_xml_str(issues: Sequence[Issue]) -> str:
    """
    Converts the list of issues into a well-formed xml string.

    Examples:

        >>> issues = [make_issue(values=10, name='i1'), make_issue(values=['a', 'b', 'c'], name='i2'),
        ... make_issue(values=(2.5, 3.5), name='i3')]
        >>> s = issues_to_xml_str(issues)
        >>> print(s.strip())
        <negotiation_template>
        <utility_space number_of_issues="3">
        <objective description="" etype="objective" index="0" name="root" type="objective">
            <issue etype="discrete" index="1" name="i1" type="discrete" vtype="integer">
                <item index="1" value="0" cost="0" description="0">
                </item>
                <item index="2" value="1" cost="0" description="1">
                </item>
                <item index="3" value="2" cost="0" description="2">
                </item>
                <item index="4" value="3" cost="0" description="3">
                </item>
                <item index="5" value="4" cost="0" description="4">
                </item>
                <item index="6" value="5" cost="0" description="5">
                </item>
                <item index="7" value="6" cost="0" description="6">
                </item>
                <item index="8" value="7" cost="0" description="7">
                </item>
                <item index="9" value="8" cost="0" description="8">
                </item>
                <item index="10" value="9" cost="0" description="9">
                </item>
            </issue>
            <issue etype="discrete" index="2" name="i2" type="discrete" vtype="discrete">
                <item index="1" value="a" cost="0" description="a">
                </item>
                <item index="2" value="b" cost="0" description="b">
                </item>
                <item index="3" value="c" cost="0" description="c">
                </item>
            </issue>
            <issue etype="real" index="3" name="i3" type="real" vtype="real">
                <range lowerbound="2.5" upperbound="3.5"></range>
            </issue>
        </objective>
        </utility_space>
        </negotiation_template>

        >>> issues2, _ = issues_from_xml_str(s)
        >>> print([_.__class__.__name__ for _ in issues2])
        ['CategoricalIssue', 'CategoricalIssue', 'ContinuousIssue']

        >>> print(len(issues2))
        3
        >>> print([str(_) for _ in issues2])
        ["i1: ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']", "i2: ['a', 'b', 'c']", 'i3: (2.5, 3.5)']
        >>> print([_.values for _ in issues2])
        [['0', '1', '2', '3', '4', '5', '6', '7', '8', '9'], ['a', 'b', 'c'], (2.5, 3.5)]
    """
    output = (
        f'<negotiation_template>\n<utility_space number_of_issues="{len(issues)}">\n'
        f'<objective description="" etype="objective" index="0" name="root" type="objective">\n'
    )

    for indx, issue in enumerate(issues):
        output += issue._to_xml_str(indx)
    output += f"</objective>\n</utility_space>\n</negotiation_template>"

    return output


def issues_to_genius(issues: Sequence[Issue], file_name: PathLike | str) -> None:
    """
    Exports a the domain issues to a GENIUS XML file.

    Args:

        issues: The issues to be exported
        file_name (str): File name to export to

    Returns:

        A tuple[Issue, ...] or dict[Issue]


    Examples:

        >>> import pkg_resources
        >>> issues, _ = issues_from_genius(file_name = pkg_resources.resource_filename('negmas'
        ...                                      , resource_name='tests/data/Laptop/Laptop-C-domain.xml'))
        >>> issues_to_genius(issues=issues, file_name = pkg_resources.resource_filename('negmas'
        ...                                    , resource_name='tests/data/LaptopConv/Laptop-C-domain.xml'))
        >>> issues2, _ = issues_from_genius(file_name = pkg_resources.resource_filename('negmas'
        ...                                    , resource_name='tests/data/LaptopConv/Laptop-C-domain.xml'))
        >>> print('\\n'.join([' '.join(list(issue.all)) for issue in issues]))
        Dell Macintosh HP
        60 Gb 80 Gb 120 Gb
        19'' LCD 20'' LCD 23'' LCD
        >>> print('\\n'.join([' '.join(list(issue.all)) for issue in issues2]))
        Dell Macintosh HP
        60 Gb 80 Gb 120 Gb
        19'' LCD 20'' LCD 23'' LCD

        - Forcing Single outcome

        >>> issues, _ = issues_from_genius(file_name = pkg_resources.resource_filename('negmas'
        ...                                      , resource_name='tests/data/Laptop/Laptop-C-domain.xml'))
        >>> print([list(issue.all) for issue in issues])
        [['Dell', 'Macintosh', 'HP'], ['60 Gb', '80 Gb', '120 Gb'], ["19'' LCD", "20'' LCD", "23'' LCD"]]

    Remarks:
        See ``from_xml_str`` for all the parameters

    """
    with open(file_name, "w") as f:
        f.write(issues_to_xml_str(issues=issues))


def issues_from_geniusweb_json(
    d: dict,
    safe_parsing=True,
    n_discretization: int | None = None,
) -> tuple[Sequence[Issue] | None, Sequence[str] | None]:
    """
    Exports a list of issues from a GeniusWeb loaded dict

    Args:

        d: the loaded dict from GeniusWeb json file
        safe_parsing (bool): Turn on extra checks
        n_discretization (Optional[int]): If not None, real valued issues are discretized with the given
        number of values
        max_cardinality (int): Maximum number of outcomes allowed (effective only if force_single_issue is True)

    Returns:

        - tuple[Issue, ...] The issues (note that issue names will be stored in the name attribute of each issue if keep_issue_names)
        - list[dict] A list of agent information dicts each contains 'agent', 'class', 'utility_file_name'

    """
    issue_values = d["issuesValues"]
    issues = [
        make_issue(name=name, values=vals["values"])
        for name, vals in issue_values.items()
    ]
    return tuple(issues), tuple()


def issues_from_geniusweb_json_str(
    json_str: str,
    safe_parsing=True,
    n_discretization: int | None = None,
) -> tuple[Sequence[Issue] | None, Sequence[str] | None]:
    """
    Exports a list/dict of issues from a GeniusWeb json file.

    Args:

        json_str (str): The string containing GENIUS style XML domain issue definitions
        safe_parsing (bool): Turn on extra checks
        n_discretization (Optional[int]): If not None, real valued issues are discretized with the given
        number of values
        max_cardinality (int): Maximum number of outcomes allowed (effective only if force_single_issue is True)

    Returns:

        - tuple[Issue, ...] The issues (note that issue names will be stored in the name attribute of each issue if keep_issue_names)
        - list[dict] A list of agent information dicts each contains 'agent', 'class', 'utility_file_name'

    """
    d = json.loads(json_str)
    return issues_from_geniusweb_json(d, safe_parsing, n_discretization)


def issues_from_xml_str(
    xml_str: str,
    safe_parsing=True,
    n_discretization: int | None = None,
) -> tuple[Sequence[Issue] | None, Sequence[str] | None]:
    """
    Exports a list/dict of issues from a GENIUS XML file.

    Args:

        xml_str (str): The string containing GENIUS style XML domain issue definitions
        safe_parsing (bool): Turn on extra checks
        n_discretization (Optional[int]): If not None, real valued issues are discretized with the given
        number of values
        max_cardinality (int): Maximum number of outcomes allowed (effective only if force_single_issue is True)

    Returns:

        - tuple[Issue, ...] The issues (note that issue names will be stored in the name attribute of each issue if keep_issue_names)
        - list[dict] A list of agent information dicts each contains 'agent', 'class', 'utility_file_name'

    Examples:

        >>> import pkg_resources
        >>> domain_file_name = pkg_resources.resource_filename('negmas'
        ...                                      , resource_name='tests/data/Laptop/Laptop-C-domain.xml')
        >>> with open(domain_file_name, 'r') as ff:
        ...     issues, _ = issues_from_xml_str(ff.read())
        >>> print([_.cardinality for _ in issues])
        [3, 3, 3]

        >>> domain_file_name = pkg_resources.resource_filename('negmas'
        ...                              , resource_name='tests/data/fuzzyagent/single_issue_domain.xml')
        >>> with open(domain_file_name, 'r') as ff:
        ...     issues, _ = issues_from_xml_str(ff.read())
        >>> len(issues)
        1
        >>> type(issues)
        <class 'tuple'>
        >>> str(issues[0]).split(': ')[-1]
        '(10.0, 40.0)'
        >>> print([_.cardinality for _ in issues])
        [inf]
    """
    root = ET.fromstring(xml_str)

    if safe_parsing and root.tag != "negotiation_template":
        raise ValueError(f"Root tag is {root.tag}: negotiation_template")

    utility_space = None
    agents = []

    for child in root:
        if child.tag == "utility_space":
            utility_space = child

            for _ in utility_space:
                if _.tag == "objective":
                    utility_space = _

                    break
        elif child.tag == "agent":
            agents.append(child.attrib)

    if utility_space is None:
        if safe_parsing:
            raise ValueError(f"No objective child was found in the root")
        utility_space = root
    issues_dict: dict[int | str, Any] = {}
    issue_info = {}

    for child in utility_space:
        if child.tag == "issue":
            indx = int(child.attrib["index"]) - 1
            issue_name = str(child.attrib["name"])
            issue_info[issue_name] = {"name": issue_name, "index": indx}
            info = {"type": "discrete", "etype": "discrete", "vtype": "discrete"}

            for a in ("type", "etype", "vtype"):
                info[a] = child.attrib.get(a, info[a])
            mytype = info["type"]

            if mytype == "discrete":
                issues_dict[issue_name] = []

                for item in child:
                    if item.tag == "item":
                        item_name = item.attrib.get("value", None)

                        if (
                            item_name not in issues_dict[issue_name]
                        ):  # ignore repeated items
                            issues_dict[issue_name].append(item_name)

            elif mytype in ("integer", "real"):
                lower_, upper_ = (
                    child.attrib.get("lowerbound", None),
                    child.attrib.get("upperbound", None),
                )

                for rng_child in child:
                    if rng_child.tag == "range":
                        lower_, upper_ = (
                            rng_child.attrib.get("lowerbound", lower_),
                            rng_child.attrib.get("upperbound", upper_),
                        )
                if lower_ is None:
                    if upper_ is not None and float(upper_) < 0:
                        lower_ = str(
                            -(sys.maxsize // 2)
                            if mytype == "integer"
                            else float("-inf")
                        )
                    else:
                        lower_ = "0"
                if upper_ is None:
                    upper_ = str(
                        (sys.maxsize // 2) if mytype == "integer" else float("-inf")
                    )
                if mytype == "integer":
                    lower, upper = int(lower_), int(upper_)
                    issues_dict[issue_name] = lower, upper
                else:
                    lower, upper = float(lower_), float(upper_)
                    if n_discretization is None:
                        issues_dict[issue_name] = lower, upper
                    else:
                        issues_dict[issue_name] = n_discretization
            else:
                # I should add the real-valued issues_dict code here
                raise ValueError(f"Unknown type: {mytype}")
        else:
            raise ValueError(f"Unknown child for objective: {child.tag}")

    issues_by_index = dict()
    for key, value in issues_dict.items():
        indx = issue_info[key]["index"]
        issues_by_index[indx] = make_issue(values=value, name=issue_info[key]["name"])
    n_issues = max(issues_by_index.keys()) + 1
    issues = []
    for i in range(n_issues):
        if i not in issues_by_index.keys():
            if safe_parsing:
                raise ValueError(
                    f"No issue with index {i} is found even though we have issues with higher indices"
                )
            issues.append(ContiguousIssue((1, 1), name=DUMMY_ISSUE_NAME))
            continue
        issues.append(issues_by_index[i])

    return tuple(issues), tuple(agents)


def issues_from_geniusweb(
    file_name: PathLike | str,
    safe_parsing=True,
    n_discretization: int | None = None,
) -> tuple[Sequence[Issue] | None, Sequence[str] | None]:
    """
    Imports a the domain issues from a GENIUS XML file.

    Args:

        file_name (str): File name to import from
        safe_parsing: Add more checks to parsing
        n_discretization: Number of discretization levels per issue

    Returns:

        A tuple of two optional lists:

            - tuple[Issue, ...] containing the issues
            - list[str] containing agent names (that are sometimes stored in the genius domain)


    Examples:

        >>> import pkg_resources
        >>> issues, _ = issues_from_genius(file_name = pkg_resources.resource_filename('negmas'
        ...                                      , resource_name='tests/data/Laptop/Laptop-C-domain.xml'))
        >>> print([_.name for _ in issues])
        ['Laptop', 'Harddisk', 'External Monitor']

    Remarks:
        See ``from_xml_str`` for all the parameters

    """
    with open(file_name, encoding="utf-8") as f:
        json_str = f.read()

        return issues_from_geniusweb_json_str(
            json_str=json_str,
            safe_parsing=safe_parsing,
            n_discretization=n_discretization,
        )


def issues_from_genius(
    file_name: PathLike | str,
    safe_parsing=True,
    n_discretization: int | None = None,
) -> tuple[Sequence[Issue] | None, Sequence[str] | None]:
    """
    Imports a the domain issues from a GENIUS XML file.

    Args:

        file_name (str): File name to import from
        safe_parsing: Add more checks to parsing
        n_discretization: Number of discretization levels per issue

    Returns:

        A tuple of two optional lists:

            - tuple[Issue, ...] containing the issues
            - list[str] containing agent names (that are sometimes stored in the genius domain)


    Examples:

        >>> import pkg_resources
        >>> issues, _ = issues_from_genius(file_name = pkg_resources.resource_filename('negmas'
        ...                                      , resource_name='tests/data/Laptop/Laptop-C-domain.xml'))
        >>> print([_.name for _ in issues])
        ['Laptop', 'Harddisk', 'External Monitor']

    Remarks:
        See ``from_xml_str`` for all the parameters

    """
    with open(file_name, encoding="utf-8") as f:
        xml_str = f.read()

        return issues_from_xml_str(
            xml_str=xml_str,
            safe_parsing=safe_parsing,
            n_discretization=n_discretization,
        )


def generate_issues(
    params: Sequence[
        int | list[str] | tuple[int, int] | Callable | tuple[float, float]
    ],
    counts: list[int] | None = None,
    names: list[str] | None = None,
) -> tuple[Issue, ...]:
    """
    Generates a set of issues with given parameters. Each is optionally repeated.

    Args:

        params: The parameters of the issues
        counts: The number of times to repeat each of the `issues`
        names: The names to assign to the issues. If None, then string representations of integers
               starting from zero will be used.

    Returns:
        list['Issue']: The list of issues with given conditions

    """
    one_each = counts is None
    int_names = names is None
    result: list[Issue] = []
    nxt = 0

    for i, issue in enumerate(params):
        count = 1 if one_each else counts[i]  # type: ignore

        for _ in range(count):
            name = str(nxt) if int_names else names[i]  # type: ignore
            # if count > 1:
            #    name = name + f' {j}'
            nxt += 1
            result.append(make_issue(values=issue, name=name))

    return tuple(result)


def discretize_and_enumerate_issues(
    issues: Iterable[Issue],
    n_discretization: int | None = 10,
    max_cardinality: int | float | None = None,
) -> list[Outcome]:
    """
    Enumerates the outcomes of a list of issues.

    Args:
        issues: The list of issues.
        max_cardinality: The maximum number of outcomes to return
    Returns:
        list of outcomes of the given type.
    """
    issues = [
        _
        if _.is_finite()
        else make_issue(values=list(_.value_generator(n_discretization)), name=_.name)
        for _ in issues
    ]
    return enumerate_issues(issues, max_cardinality=max_cardinality)


def combine_issues(
    issues: Sequence[Issue],
    name: str | None = None,
    keep_value_names=True,
    issue_sep="_",
    value_sep="-",
) -> Issue | None:
    """
    Combines multiple issues into a single issue.

    Args:
        issues: The issues to be combined
        name: The name of the resulting issue (If not given, combines input issue names)
        keep_value_names: If true, the values for the generated issue
                          will be a concatenation of values from earlier
                          issues separated by `value_sep`.
        issue_sep: Separator for the issue name (used only if `keep_issue_names`)
        value_sep: Separator for the issue name (used only if `keep_value_names`)

    Remarks:

        - Only works if the issues have finite cardinality

    """
    n_outcomes = num_outcomes(issues)
    if n_outcomes == float("inf"):
        return None
    name = issue_sep.join([_.name for _ in issues]) if name is None else name
    if keep_value_names:
        values = [
            value_sep.join([str(_) for _ in outcomes])
            for outcomes in enumerate_issues(issues)
        ]
    else:
        values = n_outcomes
    return make_issue(name=name, values=values)



from __future__ import annotations

import numbers
from random import random
from typing import Any, Generator

from negmas.helpers.types import get_full_type_name
from negmas.outcomes.base_issue import Issue
from negmas.serialization import PYTHON_CLASS_IDENTIFIER, deserialize, serialize

__all__ = ["OptionalIssue"]


class OptionalIssue(Issue):
    """
    Base class of an issues that is explicitly marked to be optional.
    Protocols can use that as they see fit. The main effect of defining an issue as optional
    is that the value `None` is allowed and returned first when enumerated (it is also counted in
    the cardinality)
    """

    def __init__(
        self,
        base: Issue,
        name: str | None = None,
    ) -> None:
        self.base = base
        self._n_values = self.base._n_values + 1
        super().__init__(values=base.values, name=name)

    @property
    def value_type(self):
        """
        Returns the type of values in this issue
        """
        return self.base._value_type

    @property
    def values(self):
        """
        Returns the raw values representation of the issue. Only use if you know what you are doing. To get all the values that can be assigned to this issue use `all` or `generate_values`
        """
        return self.base._values

    def has_limits(self) -> bool:
        """
        Checks whether the minimum and maximum values of the issue are known
        """
        return self.min_value is not None and self.max_value is not None

    def is_numeric(self) -> bool:
        return issubclass(self.base._value_type, numbers.Number)

    def is_integer(self) -> bool:
        return issubclass(self.base.value_type, numbers.Integral)

    def is_float(self) -> bool:
        return issubclass(self._value_type, numbers.Real) and not issubclass(
            self._value_type, numbers.Integral
        )

    def is_continuous(self) -> bool:
        return self.base.is_continuous()

    def is_discrete(self) -> bool:
        return not self.is_continuous()

    @property
    def cardinality(self) -> int | float:
        """Adds one to the base cardinality to handle None"""
        return self.base.cardinality + 1

    def rand(self) -> int | float | str | None:
        """Picks a random valid value."""
        if self.is_continuous():
            return self.base.rand()
        p = 1.0 - self.base.cardinality / self.cardinality
        if random() < p:
            return None
        return self.base.rand()

    @classmethod
    def from_dict(cls, d):
        """
        Constructs an issue from a dict generated using `to_dict()`
        """
        if isinstance(d, cls):
            return d
        d.pop(PYTHON_CLASS_IDENTIFIER, None)
        d["base"] = deserialize(d["base"])
        return cls(
            base=d.get("base", None),
            name=d.get("name", None),
        )

    def to_dict(self):
        """
        Converts the issue to a dictionary from which it can be constructed again using `Issue.from_dict()`
        """
        d = {PYTHON_CLASS_IDENTIFIER: get_full_type_name(type(self))}
        return dict(
            **d,
            base=serialize(self.base),
            name=self.name,
            n_values=self.cardinality + 1,
        )

    def is_valid(self, v) -> bool:
        """Checks whether the given value is valid for this issue"""
        return v is None or self.base.is_valid(v)

    def contains(self, issue: Issue) -> bool:
        """
        Checks weather this issue contains the input issue (i.e. every value in the input issue is in this issue)
        """
        return self.base.contains(issue)

    @property
    def type(self) -> str:
        """
        Returns a nice name for the issue type
        """
        return "optional_" + self.base.__class__.__name__.lower().replace("issue", "")

    def ordered_value_generator(
        self, n: int | float | None = None, grid=True, compact=False, endpoints=True
    ) -> Generator[int | None, None, None]:
        yield None
        yield from self.base.ordered_value_generator(n, grid, compact, endpoints)

    def value_generator(
        self, n: int | float | None = 10, grid=True, compact=True, endpoints=True
    ) -> Generator[Any, None, None]:
        yield None
        yield from self.base.value_generator(n, grid, compact, endpoints)

    def to_discrete(
        self, n: int | float | None = 10, grid=True, compact=True, endpoints=True
    ) -> OptionalIssue:
        return OptionalIssue(self.base.to_discrete(n, grid, compact, endpoints))

    def _to_xml_str(self, indx: int) -> str:
        # TODO: For now, we do not mark the issue as optional when saving it
        return self.base._to_xml_str(indx)

    def value_at(self, index: int):
        """
        None is assumed to be first
        """
        if index == 0:
            return None
        return self.base.value_at(index - 1)

    def rand_outcomes(
        self, n: int, with_replacement=False, fail_if_not_enough=False
    ) -> list:
        return self.base.rand_outcomes(n, with_replacement, fail_if_not_enough)

    def rand_invalid(self):
        """Pick a random *invalid* value"""
        return self.base.rand_invalid()

    @property
    def all(self) -> Generator[Any, None, None]:
        """
        A generator that generates all possible values.
        """
        yield None
        yield from self.base.all

    def __eq__(self, other):
        if not isinstance(other, OptionalIssue):
            return False
        return self.base == other.base

    def __repr__(self):
        return f"{self.__class__.__name__}({self.base}, {self.name})"

    def __str__(self):
        return f"{self.name}: {self.base}"



from __future__ import annotations

import math
import random
from abc import abstractmethod
from typing import Iterable, Sequence, TypeVar

import numpy as np

from negmas import warnings
from negmas.helpers.prob import Distribution, ScipyDistribution
from negmas.outcomes import Issue, Outcome
from negmas.outcomes.protocols import OutcomeSpace

from .base_ufun import BaseUtilityFunction, _ExtremelyDynamic

__all__ = [
    "UtilityFunction",
]

T = TypeVar("T", bound="UtilityFunction")


class UtilityFunction(_ExtremelyDynamic, BaseUtilityFunction):
    """Base for all crisp ufuns"""

    @abstractmethod
    def eval(self, offer: Outcome) -> float:
        ...

    def to_crisp(self) -> UtilityFunction:
        return self

    @classmethod
    def generate_bilateral(
        cls,
        outcomes: int | Sequence[Outcome],
        conflict_level: float = 0.5,
        conflict_delta=0.005,
    ) -> tuple[UtilityFunction, UtilityFunction]:
        """Generates a couple of utility functions

        Args:

            n_outcomes (int): number of outcomes to use
            conflict_level: How conflicting are the two ufuns to generate.
                            1.0 means maximum conflict.
            conflict_delta: How variable is the conflict at different outcomes.

        Examples:

            >>> from negmas.preferences import conflict_level
            >>> u1, u2 = UtilityFunction.generate_bilateral(outcomes=10, conflict_level=0.0
            ...                                             , conflict_delta=0.0)
            >>> print(conflict_level(u1, u2, outcomes=10))
            0.0

            >>> u1, u2 = UtilityFunction.generate_bilateral(outcomes=10, conflict_level=1.0
            ...                                             , conflict_delta=0.0)
            >>> print(conflict_level(u1, u2, outcomes=10))
            1.0

            >>> u1, u2 = UtilityFunction.generate_bilateral(outcomes=10, conflict_level=0.5
            ...                                             , conflict_delta=0.0)
            >>> 0.0 < conflict_level(u1, u2, outcomes=10) < 1.0
            True


        """
        from negmas.preferences.crisp.mapping import MappingUtilityFunction

        if isinstance(outcomes, int):
            outcomes = [(_,) for _ in range(outcomes)]
        n_outcomes = len(outcomes)
        u1 = np.random.random(n_outcomes)
        rand = np.random.random(n_outcomes)
        if conflict_level > 0.5:
            conflicting = 1.0 - u1 + conflict_delta * np.random.random(n_outcomes)
            u2 = conflicting * conflict_level + rand * (1 - conflict_level)
        elif conflict_level < 0.5:
            same = u1 + conflict_delta * np.random.random(n_outcomes)
            u2 = same * (1 - conflict_level) + rand * conflict_level
        else:
            u2 = rand

        # todo implement win_win correctly. Order the ufun then make outcomes with good outcome even better and vice
        # versa
        # u2 += u2 * win_win
        # u2 += np.random.random(n_outcomes) * conflict_delta
        u1 -= u1.min()
        u2 -= u2.min()
        u1 = u1 / u1.max()
        u2 = u2 / u2.max()
        if random.random() > 0.5:
            u1, u2 = u2, u1
        return (
            MappingUtilityFunction(dict(zip(outcomes, u1))),
            MappingUtilityFunction(dict(zip(outcomes, u2))),
        )

    @classmethod
    def generate_random_bilateral(
        cls, outcomes: int | Sequence[Outcome]
    ) -> tuple[UtilityFunction, UtilityFunction]:
        """Generates a couple of utility functions

        Args:

            n_outcomes (int): number of outcomes to use
            conflict_level: How conflicting are the two ufuns to generate. 1.0 means maximum conflict.
            conflict_delta: How variable is the conflict at different outcomes.
            zero_summness: How zero-sum like are the two ufuns.


        """
        from negmas.preferences.crisp.mapping import MappingUtilityFunction

        if isinstance(outcomes, int):
            outcomes = [(_,) for _ in range(outcomes)]
        n_outcomes = len(outcomes)
        u1 = np.random.random(n_outcomes)
        u2 = np.random.random(n_outcomes)
        u1 -= u1.min()
        u2 -= u2.min()
        u1 /= u1.max()
        u2 /= u2.max()
        return (
            MappingUtilityFunction(dict(zip(outcomes, u1))),
            MappingUtilityFunction(dict(zip(outcomes, u2))),
        )

    @classmethod
    def generate_random(
        cls,
        n: int,
        outcomes: int | Sequence[Outcome] | Iterable[Outcome],
        normalized: bool = True,
    ) -> list[UtilityFunction]:
        """Generates N mapping utility functions

        Args:
            n: number of utility functions to generate
            outcomes: number of outcomes to use
            normalized: if true, the resulting ufuns will be normlized between zero and one.


        """
        from negmas.preferences.crisp.mapping import MappingUtilityFunction

        if isinstance(outcomes, int):
            outcomes = [(_,) for _ in range(outcomes)]
        else:
            outcomes = list(outcomes)
        n_outcomes = len(outcomes)
        ufuns = []
        for _ in range(n):
            u1 = np.random.random(n_outcomes)
            if normalized:
                u1 -= u1.min()
                u1 /= u1.max()
            ufuns.append(
                MappingUtilityFunction(dict(zip(outcomes, u1)), outcomes=outcomes)
            )
        return ufuns

    def is_not_worse(self, first: Outcome, second: Outcome) -> bool:
        return self.difference(first, second) >= 0

    def difference_prob(self, first: Outcome, second: Outcome) -> Distribution:
        """
        Returns a numeric difference between the utility of the two given outcomes
        """
        return ScipyDistribution(
            loc=self.difference(first, second), scale=0.0, type="uniform"
        )

    def minmax(
        self,
        outcome_space: OutcomeSpace | None = None,
        issues: Sequence[Issue] | None = None,
        outcomes: Sequence[Outcome] | None = None,
        max_cardinality=1000,
        above_reserve=False,
    ) -> tuple[float, float]:
        """Finds the range of the given utility function for the given outcomes

        Args:
            self: The utility function
            issues: List of issues (optional)
            outcomes: A collection of outcomes (optional)
            max_cardinality: the maximum number of outcomes to try sampling (if sampling is used and outcomes are not given)
            above_reserve: If given, the minimum and maximum will be set to reserved value if they were less than it.

        Returns:
            (lowest, highest) utilities in that order

        """
        (worst, best) = self.extreme_outcomes(
            outcome_space,
            tuple(issues) if issues else issues,
            tuple(outcomes) if outcomes else outcomes,
            max_cardinality,
        )
        w, b = self(worst), self(best)
        if above_reserve:
            r = self.reserved_value
            if r is None:
                return w, b
            if b < r:
                b, w = r, r
            elif w < r:
                w = r
        return w, b

    def eval_normalized(
        self,
        offer: Outcome | None,
        above_reserve: bool = True,
        expected_limits: bool = True,
    ) -> float:
        """
        Evaluates the ufun normalizing the result between zero and one

        Args:
            offer (Outcome | None): offer
            above_reserve (bool): If True, zero corresponds to the reserved value not the minimum
            expected_limits (bool): If True, the expectation of the utility limits will be used for normalization instead of the maximum range and minimum lowest limit

        Remarks:
            - If the maximum and the minium are equal, finite and above reserve, will return 1.0.
            - If the maximum and the minium are equal, initinte or below reserve, will return 0.0.
            - For probabilistic ufuns, a distribution will still be returned.
            - The minimum and maximum will be evaluated freshly every time. If they are already cached in the ufun, the cache will be used.

        """
        r = self.reserved_value
        u = self.eval(offer) if offer else r
        mn, mx = self.minmax()
        if above_reserve:
            if mx < r:
                mx = mn = float("-inf")
            elif mn < r:
                mn = r
        d = mx - mn
        if d < 1e-5:
            warnings.warn(
                f"Ufun has equal max and min. The outcome will be normalized to zero if they were finite otherwise 1.0: {mn=}, {mx=}, {r=}, {u=}"
            )
            return 1.0 if math.isfinite(mx) else 0.0
        d = 1 / d
        return (u - mn) * d

    def __call__(self, offer: Outcome | None) -> float:
        """Calculate the utility_function value for a given outcome.

        Args:
            offer: The offer to be evaluated.


        Remarks:

            - It calls the abstract method `eval` after opationally adjusting the
              outcome type.
            - It is preferred to override eval instead of directly overriding this method
            - You cannot return None from overriden eval() functions but raise an exception (ValueError) if it was
              not possible to calculate the Value.
            - Return a float from your `eval` implementation.
            - Return the reserved value if the offer was None

        Returns:
            The utility of the given outcome
        """
        if offer is None:
            return self.reserved_value
        return self.eval(offer)

    def __getitem__(self, offer: Outcome | None) -> float | None:
        """Overrides [] operator to call the ufun allowing it to act as a mapping"""
        return self(offer)


class CrispAdapter(UtilityFunction):
    """
    Adapts any utility function to act as a crisp utility function (i.e. returning a real number)
    """

    def __init__(self, prob: BaseUtilityFunction):
        self._prob = prob

    def eval(self, offer):
        return float(self._prob.eval(offer))

    def to_stationary(self):
        return CrispAdapter(prob=self._prob.to_stationary())



from __future__ import annotations

import random
from abc import ABC, abstractmethod
from typing import Any, Generator

from negmas import warnings
from negmas.helpers import unique_name
from negmas.helpers.numeric import is_float_type, is_int_type, sample
from negmas.outcomes.base_issue import DiscreteIssue, Issue

__all__ = ["OrdinalIssue", "DiscreteOrdinalIssue"]


def generate_values(n: int) -> list[str]:
    if n > 1000_000:
        warnings.warn(
            f"You are creating an OrdinalIssue with {n} items. This is too large."
            f"Consider using something like ContiguousIssue if possible",
            warnings.NegmasMemoryWarning,
        )
    width = len(str(n))
    return list(f"{_:0{width}d}" for _ in range(n))


class OrdinalIssue(Issue, ABC):
    """
    An `Issue` that have some defined ordering of outcomes but not necessarily
    a meaningful difference function between its values.
    """

    @abstractmethod
    def ordered_value_generator(
        self, n: int = 10, grid=True, compact=False, endpoints=True
    ) -> Generator[Any, None, None]:
        ...


class DiscreteOrdinalIssue(DiscreteIssue, OrdinalIssue):
    """
    A `DiscreteIssue` that have some defined ordering of outcomes but not necessarily a meaningful difference function between its values.
    """

    def __init__(self, values, name=None) -> None:
        """
        `values` can be an integer and in this case, values will be strings
        """
        super().__init__(values, name)
        if isinstance(values, int):
            values = generate_values(values)
        else:
            values = list(values)
        types = {type(_) for _ in values}
        if len(types) == 1:
            type_ = list(types)[0]
        elif all(is_int_type(_) for _ in types):
            type_ = int
        elif all(is_float_type(_) for _ in types):
            type_ = float
        else:
            raise TypeError(
                f"Found the following types in the list of values for an "
                f"ordinal issue ({types}). Can only have one type. Try "
                f"CategoricalIssue"
            )
        self._value_type = type_
        self._n_values = len(values)
        self.min_value, self.max_value = min(values), max(values)

    def _to_xml_str(self, indx):
        output = f'    <issue etype="discrete" index="{indx + 1}" name="{self.name}" type="discrete" vtype="discrete">\n'

        for i, v in enumerate(self._values):
            output += f'        <item index="{i + 1}" value="{v}" cost="0" description="{v}">\n        </item>\n'
        output += "    </issue>\n"
        return output

    @property
    def all(self) -> Generator[Any, None, None]:
        yield from self._values  # type: ignore

    def rand_invalid(self):
        """Pick a random *invalid* value"""

        if self.is_float():
            return random.random() * self.max_value + self.max_value * 1.1

        if self.is_integer():
            return random.randint(self.max_value + 1, self.max_value * 2)

        return unique_name("") + str(random.choice(self._values)) + unique_name("")

    def ordered_value_generator(
        self, n: int = 10, grid=True, compact=False, endpoints=True
    ) -> Generator[Any, None, None]:
        """
        A generator that generates at most `n` values (in order)

        Remarks:
            - This function returns a generator for the case when the number of values is very large.
            - If you need a list then use something like:

            >>> from negmas.outcomes import make_issue
            >>> list(make_issue(5).ordered_value_generator())
            [0, 1, 2, 3, 4]
            >>> list(int(10 * _) for _ in make_issue((0.0, 1.0)).ordered_value_generator(11))
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

        """
        yield from (
            self._values[_]
            for _ in sample(
                self.cardinality, n, grid=grid, compact=compact, endpoints=endpoints
            )
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from attrs import define, field

from negmas.preferences import RankOnlyUtilityFunction
from negmas.preferences.base_ufun import BaseUtilityFunction
from negmas.preferences.mixins import StationaryMixin

from ..base import GBComponent

if TYPE_CHECKING:
    from negmas import PreferencesChange, Value
    from negmas.outcomes import Outcome

__all__ = [
    "UFunModel",
    "FrequencyUFunModel",
    "FrequencyLinearUFunModel",
    "ZeroSumModel",
]

SENTINAL = object()


class UFunModel(GBComponent, BaseUtilityFunction):
    """
    A `SAOComponent` that can model the opponent's utility function.

    Classes implementing this ufun-model, must implement the abstract `eval()`
    method to return the utility value of an outcome. They can use any callbacks
    available to `SAOComponent` to update the model.
    """


@define
class ZeroSumModel(StationaryMixin, UFunModel):
    """
    Assumes a zero-sum negotiation (i.e. $u_o$ = $-u_s$ )

    Remarks:

        - Because some negotiators do not work well with negative ufun values, we return (max - u(w)) instead of (- u(w))
    """

    above_reserve: bool = True
    rank_only: bool = False
    _effective_ufun: BaseUtilityFunction = field(init=False, default=SENTINAL)

    def on_preferences_changed(self, changes: list[PreferencesChange]):
        if not self.negotiator or not self.negotiator.ufun:
            raise ValueError("Negotiator or ufun are not known")
        self._effective_ufun = (
            self.negotiator.ufun
            if not self.rank_only
            else RankOnlyUtilityFunction(self.negotiator.ufun)
        )

    def eval(self, offer: Outcome) -> Value:
        uo = self._effective_ufun.eval_normalized(offer, self.above_reserve) * -1 + 1.0
        mn, mx = self._effective_ufun.minmax(above_reserve=False)
        return uo * (mx - mn) + mn

    def eval_normalized(
        self,
        offer: Outcome | None,
        above_reserve: bool = True,
        expected_limits: bool = True,
    ) -> Value:
        if offer is None:
            return 0.0
        return (
            self._effective_ufun.eval_normalized(offer, above_reserve, expected_limits)
            * -1
            + 1.0
        )


class FrequencyUFunModel(UFunModel):
    """
    A `PartnerUfunModel` that uses a simple frequency-based model of the opponent offers.
    """

    def eval(self, offer: Outcome) -> Value:
        raise NotImplementedError()


class FrequencyLinearUFunModel(UFunModel):
    """
    A `PartnerUfunModel` that uses a simple frequency-based model of the opponent offers assuming the ufun is `LinearAdditiveUtilityFunction` .
    """

    def eval(self, offer: Outcome) -> Value:
        raise NotImplementedError()



from __future__ import annotations

from random import choice, random
from typing import TYPE_CHECKING

from negmas.common import Distribution
from negmas.helpers.prob import ScipyDistribution

from ..mixins import StationaryMixin
from ..prob_ufun import ProbUtilityFunction

if TYPE_CHECKING:
    from negmas.outcomes import Outcome

__all__ = ["ProbRandomUtilityFunction"]


class ProbRandomUtilityFunction(StationaryMixin, ProbUtilityFunction):
    """A random utility function for a discrete outcome space"""

    def __init__(
        self,
        locs: tuple[float, float] = (0.0, 1.0),
        scales: tuple[float, float] = (0.0, 1.0),
        types: tuple[str, ...] = ("uniform",),
        *args,
        **kwargs,
    ):
        super().__init__(
            *args,
            **kwargs,
        )
        self._cache: dict[Outcome | None, tuple[float, float, str]] = dict()
        self._types = [types] if isinstance(types, str) else types
        self._loc_scale = locs[1] - locs[0]
        self._loc_offset = locs[0]
        self._scale_scale = scales[1] - scales[0]
        self._scale_offset = scales[0]

    def eval(self, offer: Outcome | None) -> Distribution:
        v = self._cache.get(offer, None)
        if v is None:
            loc = self._loc_offset + self._loc_scale * random()
            scale = self._scale_offset + self._scale_scale * random()
            typ = choice(self._types)
            self._cache[offer] = (loc, scale, typ)
        else:
            loc, scale, typ = v
        return ScipyDistribution(type=typ, loc=loc, scale=scale)



from __future__ import annotations

"""Implements GA-based negotiation mechanisms"""

import copy
import random

from attrs import define, field

from negmas.common import Action

from .mechanisms import Mechanism, MechanismState, MechanismStepResult
from .outcomes import Outcome


@define
class GAState(MechanismState):
    """Defines extra values to keep in the mechanism state. This is accessible to all negotiators"""

    dominant_outcomes: list[Outcome] = field(default=list)


class GAMechanism(Mechanism):
    """Naive GA-based mechanism that assume multi-issue discrete domains.

    Args:
        *args: positional arguments to be passed to the base Mechanism
        **kwargs: keyword arguments to be passed to the base Mechanism
        n_population: The number of outcomes for each generation
        mutate_rate: The rate of mutation
    """

    def generate(self, n: int) -> list[Outcome]:
        return self.random_outcomes(n)

    def __init__(
        self, *args, n_population: int = 100, mutate_rate: float = 0.1, **kwargs
    ):
        kwargs["initial_state"] = GAState()
        super().__init__(*args, **kwargs)
        self._current_state: GAState

        self.n_population = n_population

        self.mutate_rate = mutate_rate

        self.population = self.generate(self.n_population)

        self.dominant_outcomes = self.population[:]
        self._current_state.dominant_outcomes = self.dominant_outcomes  # type: ignore

        self.ranks = {}

    def crossover(self, outcome1: Outcome, outcome2: Outcome) -> Outcome:
        """Uniform crossover"""
        outcome = list(copy.deepcopy(outcome1))
        for i in range(len(self.issues)):
            if bool(random.getrandbits(1)):
                outcome[i] = outcome2[i]

        return tuple(outcome)

    def mutate(self, outcome: Outcome) -> Outcome:
        """Uniform crossover with random outcome"""
        return self.crossover(outcome, self.generate(1)[0])

    def select(self, outcomes: list[Outcome]) -> list[Outcome]:
        """Select Pareto optimal outcomes"""
        return self.dominant_outcomes

    def next_generation(self, parents: list[Outcome]) -> list[Outcome]:
        """Generate the next generation from parents"""
        self.population = parents[:]
        for _ in range(self.n_population - len(self.dominant_outcomes)):
            if random.random() > self.mutate_rate and len(self.dominant_outcomes) >= 2:
                self.population.append(
                    self.crossover(*random.sample(self.dominant_outcomes, 2))
                )
            else:
                self.population.append(
                    self.mutate(random.choice(self.dominant_outcomes))
                )
        return self.population

    def update_ranks(self):
        self.ranks.clear()
        outcomes = {}
        for outcome in self.population:
            outcomes[str(outcome)] = outcome  # merge duplicates
            self.ranks[str(outcome)] = {}

        # get ranking from agents
        for neg in self.negotiators:
            sorted_outcomes = list(outcomes.values())
            neg.sort(sorted_outcomes, descending=True)
            for i, outcome in enumerate(sorted_outcomes):
                self.ranks[str(outcome)][neg] = i

    def update_dominant_outcomes(self):
        """Return dominant outcomes of population"""
        self.dominant_outcomes.clear()
        outcomes = {}
        for outcome in self.population:
            outcomes[str(outcome)] = outcome  # merge duplicates

        # get dominant outcomes
        # naive approach (should be improved)
        for target in outcomes.keys():
            for outcome in outcomes.keys():
                if target == outcome:
                    continue
                for neg in self.negotiators:
                    if self.ranks[target][neg] < self.ranks[outcome][neg]:
                        break
                else:
                    break
            else:
                self.dominant_outcomes.append(outcomes[target])
        self._current_state.dominant_outcomes = self.dominant_outcomes  # type: ignore

    def __call__(
        self, state: GAState, action: Action | None = None
    ) -> MechanismStepResult:
        self.update_ranks()
        self.update_dominant_outcomes()
        self.next_generation(self.select(self.population))
        return MechanismStepResult(state=state)



"""
Base Evaluation Strategies
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from random import choice

from attrs import asdict, define

from negmas.common import MechanismState
from negmas.gb.common import GBResponse, GBState, ThreadState
from negmas.outcomes.common import Outcome

__all__ = [
    "EvaluationStrategy",
    "LocalEvaluationStrategy",
    "AnyAcceptEvaluationStrategy",
    "AllAcceptEvaluationStrategy",
    "all_accept",
    "any_accept",
]


@define
class EvaluationStrategy(ABC):
    @abstractmethod
    def __call__(
        self,
        negotiator_ids: list[str],
        state: GBState,
        history: list[GBState],
        active_thread: int | None,
    ) -> GBResponse:
        """Base class for evaluation strategies

        Args:
            negotiator_ids (list[str]): List of negotiator IDs (in the same order as threads)
            state (GBState): Current state of the mechanism
            history (list[GBState]): History of past states
            active_thread (int | None): If integer, the current thread (used for local evaluators). Global evaluators

        Returns:
            GBResponse
        """

        ...

    def __and__(self, other: EvaluationStrategy) -> EvaluationStrategy:
        return AllAcceptEvaluationStrategy([self, other])

    def __or__(self, other: EvaluationStrategy) -> EvaluationStrategy:
        return AnyAcceptEvaluationStrategy([self, other])


@define
class LocalEvaluationStrategy(EvaluationStrategy):
    def __call__(
        self,
        negotiator_ids: list[str],
        state: GBState,
        history: list[GBState],
        active_thread: int,
    ) -> GBResponse:
        """Base class for evaluation strategies

        Args:
            negotiator_ids (list[str]): List of negotiator IDs (in the same order as threads)
            state (GBState): Current state of the mechanism
            history (list[GBState]): History of past states
            active_thread (int): the current thread.

        """
        source = negotiator_ids[active_thread]
        return self.eval(
            source,
            state.threads[source],
            state.thread_history(history, source),
            state.base_state,
        )

    @abstractmethod
    def eval(
        self,
        negotiator_id: str,
        state: ThreadState,
        history: list[ThreadState],
        mechanism_state: MechanismState,
    ) -> GBResponse:
        ...


class AnyAcceptEvaluationStrategy(EvaluationStrategy):
    def __init__(self, strategies: list[EvaluationStrategy]):
        self._strategies = strategies

    def __call__(self, *args, **kwargs) -> GBResponse:
        return any_accept([_(*args, **kwargs) for _ in self._strategies])


class AllAcceptEvaluationStrategy(EvaluationStrategy):
    def __init__(self, strategies: list[EvaluationStrategy]):
        self._strategies = strategies

    def __call__(self, *args, **kwargs) -> GBResponse:
        return all_accept([_(*args, **kwargs) for _ in self._strategies])


def all_accept(responses: list[GBResponse]) -> GBResponse:
    if not responses:
        return "continue"
    if len(set(responses)) == 1:
        return responses[0]
    if any(_ is None for _ in responses):
        return None
    # we either have multiple agreements from multiple strategies or some strategies says continue
    return "continue"


def any_accept(responses: list[GBResponse]) -> GBResponse:
    if not responses:
        return "continue"
    acceptances = [_ for _ in responses if isinstance(_, Outcome)]
    if acceptances:
        return choice(acceptances)
    if any(_ is None for _ in responses):
        return None
    # we either have multiple agreements from multiple strategies or some strategies says continue
    return "continue"



"""
Negotiation tournaments created and manged as standard situated tournaments.
"""
from __future__ import annotations

from collections import defaultdict
from copy import deepcopy
from functools import partial
from itertools import chain, combinations, cycle, permutations, repeat
from os import PathLike
from random import randint
from typing import Any, Generator, Sequence

from negmas.helpers import get_class, unique_name
from negmas.inout import Scenario
from negmas.negotiators import Negotiator
from negmas.outcomes import Issue, make_issue
from negmas.preferences.crisp.linear import LinearUtilityFunction
from negmas.serialization import deserialize, serialize
from negmas.situated import Agent
from negmas.situated.neg import Condition, NegAgent, NegWorld  # , _wrap_in_agents
from negmas.tournaments.tournaments import (
    TournamentResults,
    WorldRunResults,
    create_tournament,
    tournament,
)

__all__ = [
    "random_discrete_scenarios",
    "scenarios_from_list",
    "create_neg_tournament",
    "create_cartesian_neg_tournament",
    "neg_tournament",
    "cartesian_neg_tournement",
]


AgentType = str | type[Agent] | type[Negotiator]


def neg_config_generator(
    n_competitors: int,
    scenarios: Generator[Condition, None, None],
    n_agents_per_competitor: int = 1,
    agent_names_reveal_type: bool = False,
    non_competitors: tuple[str | NegAgent] | None = None,
    non_competitor_params: tuple[dict[str, Any]] | None = None,
    compact: bool = False,
    n_repetitions: int = 1,
    **kwargs,
) -> list[dict[str, Any]]:
    """
    Used internally to generate the configuration of a world.

    Args:
        n_competitors: How many competitors will run?
        scenarios: A generator used to extract `NegScenario` objects (one config per `NegScenario`)
        n_agents_per_competitor: Must be 1
        agent_names_reveal_type: If given, the agent name will be the same as its class name
        non_competitors: Must be None
        non_competitor_params: Must be None
        compact: If given, the system tries to minimize the memory footprint and secondary storage
        kwargs: Extra arguments passed directly to the world constructor

    Returns:
        A list containing one world configuration
    """
    scenario = next(scenarios)
    if non_competitors:
        raise ValueError(
            f"Non-competitors are not supported for negotiation tournaments. (You provided {non_competitors})"
        )
    if n_agents_per_competitor != 1:
        raise ValueError(
            f"n_agents_per_competitor must be 1 ({n_agents_per_competitor} given)"
        )

    world_name = unique_name(f"{scenario.name}", add_time=False, rand_digits=1, sep=".")
    no_logs = compact
    world_params = dict(
        name=world_name,
        scenario=serialize(scenario),
        compact=compact,
        no_logs=no_logs,
    )
    world_params.update(kwargs)
    config = {
        "world_params": world_params,
        "scoring_context": {},
        "non_competitors": None,
        "non_competitor_params": None,
        "is_default": [False] * n_competitors + [True] * (len(scenario.ufuns) - 1),
        "n_repetitions": n_repetitions,
    }
    return [config]


def neg_config_assigner(
    config: list[dict[str, Any]],
    max_n_worlds: int = 1,
    n_agents_per_competitor: int = 1,
    fair: bool = True,
    competitors: Sequence[type[Negotiator] | type[Agent]] = (),
    params: Sequence[dict[str, Any]] = (),
    dynamic_non_competitors: list[type[Agent]] | None = None,
    dynamic_non_competitor_params: list[dict[str, Any]] | None = None,
    exclude_competitors_from_reassignment: bool = True,
) -> list[list[dict[str, Any]]]:
    """
    Assigns agents to the world configuration.

    In negotiation tournaments, a single assignment is generated per configuration in
    which all agents exist in the world.

    All parameters other than `competitors` and `params` are ignored
    """
    competitors, params = deepcopy(competitors), deepcopy(params)
    # competitors, params = _wrap_in_agents(competitors, params, NegAgent)
    config[0]["world_params"]["types"] = competitors
    config[0]["world_params"]["params"] = params
    n_reps = config[0].pop("n_repetitions", 1)
    if n_reps == 1:
        return [config]
    configs = []
    for i in range(n_reps):
        c = deepcopy(config[0])
        c["world_params"]["name"] += f"_{i}"
        configs.append([c])
    return configs


def neg_world_generator(**kwargs):
    """
    Generates the world
    """
    config = kwargs.pop("world_params", dict())
    config["types"], config["params"] = deepcopy(config["types"]), deepcopy(
        config["params"]
    )
    config["types"] = [get_class(_) for _ in config["types"]]
    config["scenario"] = deserialize(config["scenario"])
    # name = config["scenario"].name
    # config["name"] = name if name is not None else unique_name("world", sep="_")
    return NegWorld(**config)


def neg_score_calculator(
    worlds: list[NegWorld],
    scoring_context: dict[str, Any],
    dry_run: bool,
    scoring_method: str = "received_utility",
) -> WorldRunResults:
    """A scoring function that scores agents based on their performance.

    Args:

        worlds: The world which is assumed to be run up to the point at which the scores are to be calculated.
        scoring_context:  A dict of context parameters passed by the world generator or assigner.
        dry_run: A boolean specifying whether this is a dry_run. For dry runs, only names and types are expected in
                 the returned `WorldRunResults`
        scoring_method: the method used for scoring. Can be received_utility, partner_utility, received_advantage, partner_advantage

    Returns:
        WorldRunResults giving the names, scores, and types of factory managers.

    """
    if scoring_context is not None:
        scoring_method = scoring_context.get("scoring_method", scoring_method)
    assert len(worlds) >= 1
    result = WorldRunResults(
        world_names=[world.name for world in worlds],
        log_file_names=[world.log_file_name for world in worlds]  # type: ignore
        if worlds[0].log_file_name
        else [],
    )

    for world in worlds:

        def sumdict(a, b):
            return lambda x: 0.5 * (a(x) + b(x))

        # world = worlds
        scoring_map = dict(
            received_utility=world.received_utility,
            received_advantage=world.received_advantage,
            partner_utility=world.partner_utility,
            partner_advantage=world.partner_advantage,
            welfare=sumdict(world.received_utility, world.partner_utility),
            total_advantage=sumdict(world.received_advantage, world.partner_advantage),
        )

        extra = defaultdict(list)
        for aid, agent in world.competitors.items():
            agent_type = agent.type_name
            result.names.append(agent.name)
            result.ids.append(agent.id)
            result.types.append(agent_type)
            if dry_run:
                result.scores.append(float("nan"))
                continue
            result.scores.append(scoring_map[scoring_method](aid))  # type: ignore
            for k, f in scoring_map.items():
                # if k == scoring_method:
                #     continue
                extra[k].append(
                    dict(
                        type=agent_type,
                        score=f(aid),
                        world=world.name,
                        run_id=world.name,
                    )
                )
        for k in scoring_map.keys():
            result.extra_scores[k] = extra[k]
    return result


def _update_kwargs(kwargs, scenarios, competitors, n_repetitions):
    kwargs["config_generator"] = partial(
        neg_config_generator, scenarios=scenarios, n_repetitions=1
    )
    kwargs["n_runs_per_world"] = n_repetitions
    kwargs["config_assigner"] = neg_config_assigner
    kwargs["world_generator"] = neg_world_generator
    kwargs["score_calculator"] = neg_score_calculator
    kwargs["n_competitors_per_world"] = len(competitors)
    kwargs["n_agents_per_competitor"] = 1
    kwargs["max_worlds_per_config"] = 1
    kwargs["non_competitors"] = None
    kwargs["non_competitor_params"] = None
    kwargs["dynamic_non_competitors"] = False
    kwargs["dynamic_non_competitor_params"] = False
    kwargs["exclude_competitors_from_reassignment"] = False
    return kwargs


def create_neg_tournament(
    competitors: Sequence[str | type[Agent]],
    scenarios: Generator[Condition, None, None],
    competitor_params: Sequence[dict[str, Any]] | None = None,
    n_repetitions: int = 1,
    **kwargs,
) -> PathLike:
    """
    Creates a tournament

    Args:
        competitors: A list of class names for the competitors
        scenarios:A generator that yields `NegScenario` objects specifying negotiation scenarios upon request
        name: Tournament name
        competitor_params: A list of competitor parameters (used to initialize the competitors).
        agent_names_reveal_type: If true then the type of an agent should be readable in its name (most likely at its
                                 beginning).
        n_configs: The number of different world configs (up to competitor assignment) to be generated.
        n_runs_per_world: Number of runs per world. All of these world runs will have identical competitor assignment
                          and identical world configuration.
        total_timeout: Total timeout for the complete process
        base_tournament_path: Path at which to store all results. A new folder with the name of the tournament will be
                         created at this path. A scores.csv file will keep the scores and logs folder will keep detailed
                         logs
        parallelism: Type of parallelism. Can be 'serial' for serial, 'parallel' for parallel and 'distributed' for
                     distributed! For parallel, you can add the fraction of CPUs to use after a colon (e.g. parallel:0.5
                     to use half of the CPU in the machine). By defaults parallel uses all CPUs in the machine
        scheduler_port: Port of the dask scheduler if parallelism is dask, dist, or distributed
        scheduler_ip:   IP Address of the dask scheduler if parallelism is dask, dist, or distributed
        verbose: Verbosity
        compact: If true, compact logs will be created and effort will be made to reduce the memory footprint
        save_video_fraction: The fraction of simulations for which to save videos
        forced_logs_fraction: The fraction of simulations for which to always save logs. Notice that this has no
                              effect except if no logs were to be saved otherwise (i.e. `no_logs` is passed as True)
        video_params: The parameters to pass to the video saving function
        video_saver: The parameters to pass to the video saving function after the world
        kwargs: Arguments to pass to the `config_generator` function

    Returns:
        The path at which tournament configs are stored

    """
    return create_tournament(
        competitors=competitors,
        competitor_params=competitor_params,
        **_update_kwargs(kwargs, scenarios, competitors, n_repetitions=n_repetitions),
    )


def neg_tournament(
    competitors: list[AgentType] | tuple[AgentType, ...],
    scenarios: Generator[Condition, None, None],
    competitor_params: Sequence[dict | None] | None = None,
    n_repetitions: int = 1,
    **kwargs,
) -> TournamentResults | PathLike:
    """
    Runs a tournament

    Args:

        competitors: A list of class names for the competitors
        scenarios:A generator that yields `NegScenario` objects specifying negotiation scenarios upon request
        name: Tournament name
        competitor_params: A list of competitor parameters (used to initialize the competitors).
        stage_winners_fraction: in [0, 1).  Fraction of agents to to go to the next stage at every stage. If zero, and
                                            round_robin, it becomes a single stage competition.
        agent_names_reveal_type: If true then the type of an agent should be readable in its name (most likely at its
                                 beginning).
        n_configs: The number of different world configs (up to competitor assignment) to be generated.
        n_runs_per_world: Number of runs per world. All of these world runs will have identical competitor assignment
                          and identical world configuration.
        total_timeout: Total timeout for the complete process
        tournament_path: Path at which to store all results. A new folder with the name of the tournament will be
                         created at this path. A scores.csv file will keep the scores and logs folder will keep detailed
                         logs
        parallelism: Type of parallelism. Can be 'serial' for serial, 'parallel' for parallel and 'distributed' for
                     distributed! For parallel, you can add the fraction of CPUs to use after a colon (e.g. parallel:0.5
                     to use half of the CPU in the machine). By defaults parallel uses all CPUs in the machine
        scheduler_port: Port of the dask scheduler if parallelism is dask, dist, or distributed
        scheduler_ip:   IP Address of the dask scheduler if parallelism is dask, dist, or distributed
        world_progress_callback: A function to be called after every step of every world run (only allowed for serial
                                 and parallel evaluation and should be used with cautious).
        tournament_progress_callback: A function to be called with `WorldRunResults` after each world finished
                                      processing
        verbose: Verbosity
        configs_only: If true, a config file for each
        compact: If true, compact logs will be created and effort will be made to reduce the memory footprint
        print_exceptions: If true, print all exceptions to screen
        metric: The metric to use for evaluation
        save_video_fraction: The fraction of simulations for which to save videos
        forced_logs_fraction: The fraction of simulations for which to always save logs. Notice that this has no
                              effect except if no logs were to be saved otherwise (i.e. `no_logs` is passed as True)
        video_params: The parameters to pass to the video saving function
        video_saver: The parameters to pass to the video saving function after the world
        max_attempts: The maximum number of times to retry running simulations
        extra_scores_to_use: The type of extra-scores to use. If None normal scores will be used. Only effective if scores is None.
        kwargs: Arguments to pass to the `config_generator` function

    Returns:
        `TournamentResults` The results of the tournament or a `PathLike` giving the location where configs were saved

    """
    return tournament(
        competitors=competitors,  # type: ignore
        competitor_params=competitor_params,  # type: ignore
        **_update_kwargs(kwargs, scenarios, competitors, n_repetitions=n_repetitions),
    )


def random_discrete_scenarios(
    issues: list[Issue | int | tuple[int, int]],
    partners: list[type[Negotiator]] | tuple[type[Negotiator], ...],
    n_negotiators=2,
    positions: int | tuple[int, int] | None = None,
    normalized=True,
    ufun_type=LinearUtilityFunction,
    roles: tuple[str, ...] | None = None,
    partner_extraction_method="round-robin",
) -> Generator[Condition, None, None]:
    """
    Generates an infinite sequence of random discrete scenarios

    Args:
        issues: A list defining the issue space. Each element can be an `Issue`
                object, an integer (defining the number of outcomes) or a tuple
                of two integers (defining the minimum and maximum number of outcomes
                for the issue).
        partners: A list of `Negotiator` types from which partners are extracted.
                  The system will extract `n_negotiators` - 1 partners for each
                  scenario.
        n_negotiators: The number of negotiators in each negotiation.
        positions: The positions at which the competitors will be added in all
                   negotiations.
        normalized: Will the ufuns generated by normalized
        ufun_type: Type of the utility function to use.
        roles: The roles of the `n_negotiators` (including the competitor) in negotiations
        partner_extraction_method: The method used to create partners for negotaitions
                                   from the given `partners` list:

                                   - round-robin: will extract overalapping `n_negotiators` - 1 sets
                                   - permutations: Will use all `n_negotaitors` - 1 permutations
                                   - random: Will sample randm sets of `n_negotiators` - 1 partners
                                   - compinations: Will use all `n_negotiators` - 1 combinations
    """
    partners = list(partners)
    if positions is None:
        positions = (0, n_negotiators)
    elif isinstance(positions, int):
        positions = (positions, positions)

    while len(partners) < n_negotiators - 1:
        partners += [_ for _ in partners]

    def intin(i):
        if isinstance(i, int):
            return i
        return randint(i[0], i[1])

    while True:
        current_issues = [
            _ if isinstance(_, Issue) else make_issue(values=intin(_), name=f"i{i}")
            for i, _ in enumerate(issues)
        ]
        ufuns = [
            ufun_type.random(
                current_issues, reserved_value=(0.0, 0.2), normalized=normalized
            )
            for _ in range(n_negotiators)
        ]

        def partners_generator():
            if partner_extraction_method.startswith("permutation"):
                yield from permutations(partners, n_negotiators - 1)
            if partner_extraction_method.startswith("combination"):
                yield from combinations(partners, n_negotiators - 1)
            else:
                is_random = partner_extraction_method.startswith("random")
                n = len(partners) - n_negotiators + 1
                for j in range(n):
                    if is_random:
                        yield partners[randint(0, n) : j + n_negotiators - 1]
                    else:
                        yield partners[j : j + n_negotiators - 1]

        for u in permutations(ufuns):
            for index in range(*positions):
                for p in partners_generator():
                    assert len(u) == len(p) + 1
                    yield Condition(
                        name="d0",
                        ufuns=u,
                        issues=current_issues,  # type: ignore
                        partner_types=tuple(p),
                        index=index,
                        roles=roles,
                    )


def scenarios_from_list(
    scenarios: list[Condition],
) -> Generator[Condition, None, None]:
    """
    Creates an appropriate `NegScenario` generator from a list/tuple of scenarios
    """
    while True:
        yield from cycle(scenarios)


def _make_negs(
    scenarios,
    competitors,
    competitor_params,
    non_competitors,
    non_competitor_params,
    rotate_ufuns=False,
):
    competitors = list(competitors)
    if competitor_params is None:
        competitor_params = [dict() for _ in range(len(competitors))]
    else:
        competitor_params = list(competitor_params)
        assert len(competitor_params) == len(competitors)

    if non_competitors is None:
        non_competitors = tuple()
    non_competitors = list(non_competitors)
    if non_competitor_params is None:
        non_competitor_params = [dict() for _ in range(len(non_competitors))]
    else:
        non_competitor_params = list(non_competitor_params)
        assert len(non_competitor_params) == len(non_competitors)

    try:
        intersection = set(non_competitors).intersection(set(competitors))
        assert (
            not intersection
        ), f"Non-competitors and competitors must be disjoint. This is the intersection between them now: {intersection}"
    except:
        pass

    def make_neg_scenarios(
        scenarios: tuple[Scenario, ...] | list[Scenario]
    ) -> list[Condition]:
        negs = []
        for s in scenarios:
            k = 0
            assert (
                len(s.ufuns) == 2
            ), f"Only supporting bilateral negotiations: Scenario {s.outcome_space.name} has {len(s.ufuns)} ufuns"
            for typ, params, score in chain(
                zip(competitors, competitor_params, repeat(True)),
                zip(non_competitors, competitor_params, repeat(False)),
            ):
                ufuns = [list(s.ufuns)]
                if rotate_ufuns:
                    for _ in range(len(ufuns) - 1):
                        u = ufuns[-1]
                        ufuns.append([u[-1]] + u[:-1])
                indices = [0, 1] if not score else [0]
                for indx in indices:
                    for u in ufuns:
                        negs.append(
                            Condition(
                                name=f"{s.outcome_space.name}_{k}"
                                if s.outcome_space.name
                                else unique_name("s"),
                                issues=s.outcome_space.issues,
                                ufuns=tuple(u),
                                partner_types=(get_class(typ),),
                                partner_params=(params,),
                                scored_indices=tuple(range(len(s.ufuns)))
                                if score
                                else None,
                                index=indx,
                            )
                        )
                        k += 1

        return negs

    neg_scenarios = make_neg_scenarios(scenarios)
    return (
        neg_scenarios,
        competitors,
        competitor_params,
        non_competitors,
        non_competitor_params,
    )


def create_cartesian_neg_tournament(
    competitors: list[AgentType] | tuple[AgentType, ...],
    scenarios: list[Scenario] | tuple[Scenario, ...],
    competitor_params: Sequence[dict[str, None] | None] | None = None,
    non_competitors: list[AgentType] | tuple[AgentType, ...] = tuple(),
    non_competitor_params: Sequence[dict | None] | None = None,
    rotate_ufuns: bool = False,
    n_repetitions: int = 1,
    n_steps: int = 2,
    n_rounds: int | None = None,
    timelimit: float | None = None,
    **kwargs,
) -> PathLike:
    """
    Creates a Cartesian tournament (every competitor against every other competitor)

    Args:

        competitors: A list of class names for the competitors.
        non_competitors: A list of class names for agents that will run against the competitors but never be evaluated themselves.
        scenarios:A generator that yields `NegScenario` objects specifying negotiation scenarios upon request
        name: Tournament name
        competitor_params: A list of competitor parameters (used to initialize the competitors).
        non_competitor_params: A list of non-competitor parameters (used to initialize the non-competitors).
        rotate_ufuns: If `True`, all N rotations of the N ufuns in every scenario will be tried.
        stage_winners_fraction: in [0, 1).  Fraction of agents to to go to the next stage at every stage. If zero, and
                                            round_robin, it becomes a single stage competition.
        agent_names_reveal_type: If true then the type of an agent should be readable in its name (most likely at its
                                 beginning).
        n_configs: The number of different world configs (up to competitor assignment) to be generated.
        n_runs_per_world: Number of runs per world. All of these world runs will have identical competitor assignment
                          and identical world configuration.
        total_timeout: Total timeout for the complete process
        tournament_path: Path at which to store all results. A new folder with the name of the tournament will be
                         created at this path. A scores.csv file will keep the scores and logs folder will keep detailed
                         logs
        parallelism: Type of parallelism. Can be 'serial' for serial, 'parallel' for parallel and 'distributed' for
                     distributed! For parallel, you can add the fraction of CPUs to use after a colon (e.g. parallel:0.5
                     to use half of the CPU in the machine). By defaults parallel uses all CPUs in the machine
        scheduler_port: Port of the dask scheduler if parallelism is dask, dist, or distributed
        scheduler_ip:   IP Address of the dask scheduler if parallelism is dask, dist, or distributed
        world_progress_callback: A function to be called after every step of every world run (only allowed for serial
                                 and parallel evaluation and should be used with cautious).
        tournament_progress_callback: A function to be called with `WorldRunResults` after each world finished
                                      processing
        verbose: Verbosity
        configs_only: If true, a config file for each
        compact: If true, compact logs will be created and effort will be made to reduce the memory footprint
        print_exceptions: If true, print all exceptions to screen
        metric: The metric to use for evaluation
        forced_logs_fraction: The fraction of simulations for which to always save logs. Notice that this has no
                              effect except if no logs were to be saved otherwise (i.e. `no_logs` is passed as True)
        save_video_fraction: The fraction of simulations for which to save videos
        video_params: The parameters to pass to the video saving function
        video_saver: The parameters to pass to the video saving function after the world
        max_attempts: The maximum number of times to retry running simulations
        extra_scores_to_use: The type of extra-scores to use. If None normal scores will be used. Only effective if scores is None.
        kwargs: Arguments to pass to the `config_generator` function

    Returns:
        `TournamentResults` The results of the tournament or a `PathLike` giving the location where configs were saved

    """
    kwargs["n_steps"] = n_steps
    if n_rounds is not None:
        kwargs["neg_n_steps"] = n_rounds
    if timelimit is not None:
        kwargs["neg_time_limit"] = n_rounds
    (
        neg_scenarios,
        competitors,
        competitor_params,
        non_competitors,
        non_competitor_params,
    ) = _make_negs(
        scenarios,
        competitors,
        competitor_params,
        non_competitors,
        non_competitor_params,
        rotate_ufuns=rotate_ufuns,
    )
    return create_neg_tournament(
        competitors=competitors,  # type: ignore
        scenarios=scenarios_from_list(neg_scenarios),
        competitor_params=competitor_params,  # type: ignore
        n_repetitions=n_repetitions,
        **kwargs,
    )


def cartesian_neg_tournement(
    competitors: list[AgentType] | tuple[AgentType, ...],
    scenarios: list[Scenario] | tuple[Scenario, ...],
    competitor_params: Sequence[dict | None] | None = None,
    non_competitors: list[AgentType] | tuple[AgentType, ...] = tuple(),
    non_competitor_params: Sequence[dict | None] | None = None,
    rotate_ufuns=False,
    n_repetitions: int = 1,
    n_steps: int = 2,
    n_rounds: int | None = None,
    timelimit: float | None = None,
    **kwargs,
) -> TournamentResults | PathLike:
    """
    Runs a Cartesian tournament between `competitors` optionally rotating preferences.

    Args:

        competitors: A list of class names for the competitors.
        non_competitors: A list of class names for agents that will run against the competitors but never be evaluated themselves.
        scenarios:A generator that yields `NegScenario` objects specifying negotiation scenarios upon request
        n_repetitions: Number of time each world with a single negotiation scenario is repeated
        timelimit: Time limit per negotiation
        n_rounds: Number of rounds per negotiation
        name: Tournament name
        competitor_params: A list of competitor parameters (used to initialize the competitors).
        non_competitor_params: A list of non-competitor parameters (used to initialize the non-competitors).
        stage_winners_fraction: in [0, 1).  Fraction of agents to to go to the next stage at every stage. If zero, and
                                            round_robin, it becomes a single stage competition.
        agent_names_reveal_type: If true then the type of an agent should be readable in its name (most likely at its
                                 beginning).
        total_timeout: Total timeout for the complete process
        tournament_path: Path at which to store all results. A new folder with the name of the tournament will be
                         created at this path. A scores.csv file will keep the scores and logs folder will keep detailed
                         logs
        parallelism: Type of parallelism. Can be 'serial' for serial, 'parallel' for parallel and 'distributed' for
                     distributed! For parallel, you can add the fraction of CPUs to use after a colon (e.g. parallel:0.5
                     to use half of the CPU in the machine). By defaults parallel uses all CPUs in the machine
        scheduler_port: Port of the dask scheduler if parallelism is dask, dist, or distributed
        scheduler_ip:   IP Address of the dask scheduler if parallelism is dask, dist, or distributed
        world_progress_callback: A function to be called after every step of every world run (only allowed for serial
                                 and parallel evaluation and should be used with cautious).
        tournament_progress_callback: A function to be called with `WorldRunResults` after each world finished
                                      processing
        verbose: Verbosity
        configs_only: If true, a config file for each
        compact: If true, compact logs will be created and effort will be made to reduce the memory footprint
        print_exceptions: If true, print all exceptions to screen
        metric: The metric to use for evaluation
        forced_logs_fraction: The fraction of simulations for which to always save logs. Notice that this has no
                              effect except if no logs were to be saved otherwise (i.e. `no_logs` is passed as True)
        save_video_fraction: The fraction of simulations for which to save videos
        video_params: The parameters to pass to the video saving function
        video_saver: The parameters to pass to the video saving function after the world
        max_attempts: The maximum number of times to retry running simulations
        extra_scores_to_use: The type of extra-scores to use. If None normal scores will be used. Only effective if scores is None.
        kwargs: Arguments to pass to the `config_generator` function

    Returns:
        `TournamentResults` The results of the tournament or a `PathLike` giving the location where configs were saved

    """
    kwargs["n_steps"] = n_steps
    if n_rounds is not None:
        kwargs["neg_n_steps"] = n_rounds
    if timelimit is not None:
        kwargs["neg_time_limit"] = n_rounds
    (
        neg_scenarios,
        competitors,
        competitor_params,
        non_competitors,
        non_competitor_params,
    ) = _make_negs(
        scenarios,
        competitors,
        competitor_params,
        non_competitors,
        non_competitor_params,
        rotate_ufuns=rotate_ufuns,
    )
    kwargs["n_configs"] = len(neg_scenarios)
    return neg_tournament(
        competitors=competitors,
        scenarios=scenarios_from_list(neg_scenarios),
        competitor_params=competitor_params,
        n_repetitions=n_repetitions,
        **kwargs,
    )


if __name__ == "__main__":
    from negmas.sao import AspirationNegotiator, NaiveTitForTatNegotiator

    scenarios = random_discrete_scenarios(
        issues=[5, 4, (3, 5)],
        partners=[AspirationNegotiator, NaiveTitForTatNegotiator],
    )
    print(
        neg_tournament(
            n_configs=2 * 2 * 2 * 4,
            scenarios=scenarios,
            competitors=[AspirationNegotiator, NaiveTitForTatNegotiator],
            n_steps=1,
            neg_n_steps=10,
            neg_time_limit=None,
            name="neg-tour-test",
        )
    )



from __future__ import annotations

import random
from typing import Any, Callable

from negmas.common import MechanismState, NegotiatorMechanismInterface
from negmas.helpers import get_class
from negmas.helpers.numeric import make_range
from negmas.outcomes import Issue, Outcome
from negmas.serialization import PYTHON_CLASS_IDENTIFIER, deserialize, serialize

from .base import Value
from .base_ufun import BaseUtilityFunction
from .mixins import StateDependentUFunMixin

__all__ = ["LinDiscountedUFun", "ExpDiscountedUFun", "DiscountedUtilityFunction"]


class DiscountedUtilityFunction(StateDependentUFunMixin, BaseUtilityFunction):
    """Base class for all discounted ufuns"""

    def __init__(self, ufun: BaseUtilityFunction, **kwargs):
        super().__init__(**kwargs)
        self.ufun = ufun

    def is_state_dependent(self):
        return True

    def to_stationary(self):
        return self.ufun.to_stationary()


class ExpDiscountedUFun(DiscountedUtilityFunction):
    """A discounted utility function based on some factor of the negotiation

    Args:
        ufun: The utility function that is being discounted
        discount: discount factor
        factor: str -> The name of the AgentMechanismInterface variable based on which discounting operate
        callable -> must receive a mechanism info object and returns a float representing the factor

    """

    def __init__(
        self,
        ufun: BaseUtilityFunction,
        discount: float | None = None,
        factor: str | Callable[[MechanismState], float] = "step",
        name=None,
        reserved_value: Value = float("-inf"),
        dynamic_reservation=True,
        id=None,
        **kwargs,
    ):
        super().__init__(
            ufun=ufun, name=name, reserved_value=reserved_value, id=id, **kwargs
        )
        self.ufun = ufun
        self.discount = discount
        self.factor = factor
        self.dynamic_reservation = dynamic_reservation

    def minmax(
        self,
        outcome_space=None,
        issues=None,
        outcomes=None,
        max_cardinality=10_000,
        above_reserve=False,
    ) -> tuple[float, float]:
        return self.ufun.minmax(
            outcome_space,
            issues,
            outcomes,
            max_cardinality,
            above_reserve=above_reserve,
        )

    def shift_by(self, offset: float, shift_reserved: bool = True) -> ExpDiscountedUFun:
        return ExpDiscountedUFun(
            outcome_space=self.outcome_space,
            ufun=self.ufun.shift_by(offset, shift_reserved),
            discount=self.discount,
            factor=self.factor,
            name=self.name,
            reserved_value=self.reserved_value + offset
            if shift_reserved
            else self.reserved_value,
            dynamic_reservation=self.dynamic_reservation,
        )

    def scale_by(self, scale: float, scale_reserved: bool = True) -> ExpDiscountedUFun:
        return ExpDiscountedUFun(
            outcome_space=self.outcome_space,
            ufun=self.ufun.scale_by(scale, scale_reserved),
            discount=self.discount,
            factor=self.factor,
            name=self.name,
            reserved_value=self.reserved_value + scale
            if scale_reserved
            else self.reserved_value,
            dynamic_reservation=self.dynamic_reservation,
        )

    def to_dict(self) -> dict[str, Any]:
        d = super().to_dict()
        return dict(
            **d,
            ufun=serialize(self.ufun),
            discount=self.discount,
            factor=self.factor,
            dynamic_reservation=self.dynamic_reservation,
        )

    @classmethod
    def from_dict(cls, d: dict[str, Any]):
        d.pop(PYTHON_CLASS_IDENTIFIER, None)
        d["ufun"] = deserialize(d["ufun"])
        return cls(**d)

    @classmethod
    def random(
        cls,
        issues,
        reserved_value=(0.0, 1.0),
        normalized=True,
        discount_range=(0.8, 1.0),
        base_preferences_type: (
            str | type[BaseUtilityFunction]
        ) = "negmas.LinearAdditiveUtilityFunction",
        **kwargs,
    ) -> ExpDiscountedUFun:
        """Generates a random ufun of the given type"""
        reserved_value = make_range(reserved_value)
        discount_range = make_range(discount_range)
        kwargs["discount"] = (
            random.random() * (discount_range[1] - discount_range[0])
            + discount_range[0]
        )
        kwargs["reserved_value"] = (
            random.random() * (reserved_value[1] - reserved_value[0])
            + reserved_value[0]
        )
        return cls(
            get_class(base_preferences_type).random(  # type: ignore
                issues, reserved_value=reserved_value, normalized=normalized
            ),
            **kwargs,
        )

    # @lru_cache(100)
    # def eval_normalized(self, offer: Outcome | None, above_reserve: bool = True, expected_limits: bool = True) -> Value:
    #     """
    #     Caches the top 100 values as the  normalized value for exponentially discounted ufun does not change with time.
    #     """
    #     return super().eval_normalized(offer, above_reserve, expected_limits)

    def eval_on_state(
        self,
        offer: Outcome,
        nmi: NegotiatorMechanismInterface | None = None,
        state: MechanismState | None = None,
    ):
        if offer is None and not self.dynamic_reservation:
            return self.reserved_value
        u = self.ufun(offer)
        if not self.discount or self.discount == 1.0 or state is None:
            return u
        if isinstance(self.factor, str):
            factor = getattr(state, self.factor)
        else:
            factor = self.factor(state)
        return (self.discount**factor) * u

    def xml(self, issues: list[Issue]) -> str:
        if not hasattr(self.ufun, "xml"):
            raise ValueError(
                f"Cannot serialize because my internal ufun of type {self.ufun.type} is not serializable"
            )
        output = self.ufun.xml(issues)  # type: ignore
        output += "</objective>\n"
        factor = None
        if self.factor is not None:
            factor = str(self.factor)
        if self.discount is not None:
            output += f'<discount_factor value="{self.discount}" '
            if factor is not None and factor != "step":
                output += f' variable="{factor}" '
            output += "/>\n"
        return output

    @property
    def base_type(self):
        return self.ufun.type

    @property
    def type(self):
        return self.ufun.type + "_exponentially_discounted"

    def __getattr__(self, item):
        return getattr(self.ufun, item)

    def __str__(self):
        return f"{self.ufun.type}-cost:{self.discount} based on {self.factor}"


class LinDiscountedUFun(DiscountedUtilityFunction):
    """A utility function with linear discounting based on some factor of the negotiation

    Args:

        ufun: The utility function that is being discounted
        cost: discount factor
        factor: str -> The name of the AgentMechanismInterface variable based on which discounting operate
        callable -> must receive a mechanism info object and returns a float representing the factor
        power: A power to raise the total cost to before discounting it from the utility_function value

    """

    def __init__(
        self,
        ufun: BaseUtilityFunction,
        cost: float | None = None,
        factor: str | Callable[[MechanismState], float] = "current_step",
        power: float | None = 1.0,
        name=None,
        reserved_value: Value = float("-inf"),
        dynamic_reservation=True,
        id=None,
        **kwargs,
    ):
        super().__init__(
            ufun=ufun, name=name, reserved_value=reserved_value, id=id, **kwargs
        )
        if power is None:
            power = 1.0
        self.ufun = ufun
        self.cost = cost
        self.factor = factor
        self.power = power
        self.dynamic_reservation = dynamic_reservation

    def to_dict(self) -> dict[str, Any]:
        d = super().to_dict()
        return dict(
            **d,
            ufun=serialize(self.ufun),
            cost=self.cost,
            power=self.power,
            factor=self.factor,
            dynamic_reservation=self.dynamic_reservation,
        )

    @classmethod
    def from_dict(cls, d: dict[str, Any]):
        d.pop(PYTHON_CLASS_IDENTIFIER, None)
        d["ufun"] = deserialize(d["ufun"])
        return cls(**d)

    def eval_on_state(
        self,
        offer: Outcome,
        nmi: NegotiatorMechanismInterface | None = None,
        state: MechanismState | None = None,
    ):
        if offer is None and not self.dynamic_reservation:
            return self.reserved_value
        u = self.ufun(offer)
        if not self.cost or self.cost == 0.0 or state is None:
            return u
        if isinstance(self.factor, str):
            factor = getattr(state, self.factor)
        else:
            factor = self.factor(state)
        return u - ((factor * self.cost) ** self.power)

    def xml(self, issues: list[Issue]) -> str:
        if not hasattr(self.ufun, "xml"):
            raise ValueError(
                f"Cannot serialize because my internal ufun of type {self.ufun.type} is not serializable"
            )
        output = self.ufun.xml(issues)  # type: ignore
        output += "</objective>\n"
        factor = None
        if self.factor is not None:
            factor = str(self.factor)
        if self.cost is not None:
            output += f'<cost value="{self.cost}" '
            if factor is not None and factor != "step":
                output += f' variable="{factor}" '
            if self.power is not None and self.power != 1.0:
                output += f' power="{self.power}" '
            output += "/>\n"

        return output

    @property
    def base_type(self):
        return self.ufun.type

    @property
    def type(self):
        return self.ufun.type + "_linearly_discounted"

    @classmethod
    def random(
        cls,
        issues,
        reserved_value=(0.0, 1.0),
        normalized=True,
        cost_range=(0.8, 1.0),
        power_range=(0.0, 1.0),
        base_preferences_type: type[BaseUtilityFunction]
        | str = "negmas.LinearAdditiveUtilityFunction",
        **kwargs,
    ) -> LinDiscountedUFun:
        """Generates a random ufun of the given type"""
        reserved_value = make_range(reserved_value)
        cost_range = make_range(cost_range)
        power_range = make_range(power_range)
        kwargs["cost"] = (
            random.random() * (cost_range[1] - cost_range[0]) + cost_range[0]
        )
        kwargs["power"] = (
            random.random() * (power_range[1] - power_range[0]) + power_range[0]
        )
        kwargs["reserved_value"] = (
            random.random() * (reserved_value[1] - reserved_value[0])
            + reserved_value[0]
        )
        return cls(
            get_class(base_preferences_type).random(  # type: ignore
                issues, reserved_value=kwargs["reserved_value"], normalized=normalized
            ),
            **kwargs,
        )

    def __getattr__(self, item):
        return getattr(self.ufun, item)

    def __str__(self):
        return f"{self.ufun.type}-cost:{self.cost} raised to {self.power} based on {self.factor}"

    def minmax(
        self,
        outcome_space=None,
        issues=None,
        outcomes=None,
        max_cardinality=10_000,
        above_reserve=False,
    ) -> tuple[float, float]:
        return self.ufun.minmax(
            outcome_space,
            issues,
            outcomes,
            max_cardinality,
            above_reserve=above_reserve,
        )



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from negmas.gb.common import get_offer

from ....negotiators.modular import ModularNegotiator
from ..base import GBNegotiator

if TYPE_CHECKING:
    from ....outcomes import Outcome
    from ...common import ResponseType
    from ...components import GBComponent

if TYPE_CHECKING:
    from negmas.gb import GBState

__all__ = ["GBModularNegotiator"]


class GBModularNegotiator(ModularNegotiator, GBNegotiator):
    """
    A generic modular GB negotiator.
    """

    _components: list[GBComponent]

    def components(self) -> tuple[GBComponent, ...]:
        return super().components  # type: ignore

    @abstractmethod
    def generate_response(
        self, state: GBState, offer: Outcome | None, source: str | None = None
    ) -> ResponseType:
        ...

    @abstractmethod
    def generate_proposal(self, state: GBState) -> Outcome | None:
        ...

    def propose(self, state: GBState) -> Outcome | None:
        for c in self._components:
            c.before_proposing(state)
        offer = self.generate_proposal(state)
        for c in self._components:
            c.after_proposing(state, offer=offer)
        return offer

    def respond(self, state: GBState, source: str | None = None) -> ResponseType:
        offer = get_offer(state, source)
        for c in self._components:
            c.before_responding(state=state, offer=offer, source=source)
        response = self.generate_response(state=state, offer=offer, source=source)
        for c in self._components:
            c.after_responding(
                state=state, offer=offer, response=response, source=source
            )
        return response

    def on_partner_joined(self, partner: str):
        """
        Called when a partner joins the negotiation.

        This is only receivd if the mechanism is sending notifications.
        """
        for c in self._components:
            c.on_partner_joined(partner)

    def on_partner_left(self, partner: str):
        """
        Called when a partner leaves the negotiation.

        This is only receivd if the mechanism is sending notifications.
        """
        for c in self._components:
            c.on_partner_left(partner)

    def on_partner_ended(self, partner: str):
        """
        Called when a partner ends the negotiation.

        Note  that the negotiator owning this component may never receive this
        offer.
        This is only receivd if the mechanism is sending notifications
        on every offer.
        """
        for c in self._components:
            c.on_partner_ended(partner)

    def on_partner_proposal(
        self, state: GBState, partner_id: str, offer: Outcome
    ) -> None:
        """
        A callback called by the mechanism when a partner proposes something

        Args:
            state: `GBState` giving the state of the negotiation when the offer was porposed.
            partner_id: The ID of the agent who proposed
            offer: The proposal.

        Remarks:
            - Will only be called if `enable_callbacks` is set for the mechanism
        """
        for c in self._components:
            c.on_partner_proposal(state=state, partner_id=partner_id, offer=offer)

    def on_partner_response(
        self,
        state: GBState,
        partner_id: str,
        outcome: Outcome,
        response: ResponseType,
    ) -> None:
        """
        A callback called by the mechanism when a partner responds to some offer

        Args:
            state: `GBState` giving the state of the negotiation when the partner responded.
            partner_id: The ID of the agent who responded
            outcome: The proposal being responded to.
            response: The response

        Remarks:
            - Will only be called if `enable_callbacks` is set for the mechanism
        """
        for c in self._components:
            c.on_partner_response(
                state=state,
                partner_id=partner_id,
                outcome=outcome,
                response=response,
            )



from __future__ import annotations

import numpy as np
import pytest

from negmas.outcomes import ContinuousInfiniteIssue, CountableInfiniteIssue, make_os
from negmas.outcomes.base_issue import make_issue
from negmas.outcomes.callable_issue import CallableIssue
from negmas.outcomes.cardinal_issue import DiscreteCardinalIssue
from negmas.outcomes.categorical_issue import CategoricalIssue
from negmas.outcomes.contiguous_issue import ContiguousIssue
from negmas.outcomes.continuous_issue import ContinuousIssue
from negmas.outcomes.optional_issue import OptionalIssue
from negmas.outcomes.ordinal_issue import DiscreteOrdinalIssue, generate_values


def test_make_issue_generation():
    assert isinstance(make_issue((0, 5)), ContiguousIssue)
    assert isinstance(make_issue(10), ContiguousIssue)
    assert isinstance(make_issue([1, 2, 3, 5]), DiscreteCardinalIssue)
    assert isinstance(make_issue(["a", "b", "c"]), CategoricalIssue)
    assert isinstance(make_issue(lambda: 1), CallableIssue)
    assert isinstance(make_issue((0.0, 5.0)), ContinuousIssue)
    assert isinstance(make_issue((0, float("inf"))), CountableInfiniteIssue)
    assert isinstance(make_issue((float("-inf"), 0)), CountableInfiniteIssue)
    assert isinstance(make_issue((0.0, float("inf"))), ContinuousInfiniteIssue)
    assert isinstance(
        make_issue((float("-inf"), float("inf"))), ContinuousInfiniteIssue
    )
    assert isinstance(make_issue((float("-inf"), 5.0)), ContinuousInfiniteIssue)
    issue = make_issue((0.0, float("inf")), optional=True)
    assert isinstance(issue, OptionalIssue) and isinstance(
        issue.base, ContinuousInfiniteIssue
    )
    issue = make_issue((0, 5), optional=True)
    assert isinstance(issue, OptionalIssue) and isinstance(issue.base, ContiguousIssue)
    issue = make_issue(10, optional=True)
    assert isinstance(issue, OptionalIssue) and isinstance(issue.base, ContiguousIssue)
    issue = make_issue([1, 2, 3, 5], optional=True)
    assert isinstance(issue, OptionalIssue) and isinstance(
        issue.base, DiscreteCardinalIssue
    )
    issue = make_issue(["a", "b", "c"], optional=True)
    assert isinstance(issue, OptionalIssue) and isinstance(issue.base, CategoricalIssue)
    issue = make_issue(lambda: 1, optional=True)
    assert isinstance(issue, OptionalIssue) and isinstance(issue.base, CallableIssue)
    issue = make_issue((0.0, 5.0), optional=True)
    assert isinstance(issue, OptionalIssue) and isinstance(issue.base, ContinuousIssue)
    issue = make_issue((0, float("inf")), optional=True)
    assert isinstance(issue, OptionalIssue) and isinstance(
        issue.base, CountableInfiniteIssue
    )
    issue = make_issue((float("-inf"), 0), optional=True)
    assert isinstance(issue, OptionalIssue) and isinstance(
        issue.base, CountableInfiniteIssue
    )
    issue = make_issue((0.0, float("inf")), optional=True)
    assert isinstance(issue, OptionalIssue) and isinstance(
        issue.base, ContinuousInfiniteIssue
    )
    issue = make_issue((float("-inf"), float("inf")), optional=True)
    assert isinstance(issue, OptionalIssue) and isinstance(
        issue.base, ContinuousInfiniteIssue
    )
    issue = make_issue((float("-inf"), 5.0), optional=True)
    assert isinstance(issue, OptionalIssue) and isinstance(
        issue.base, ContinuousInfiniteIssue
    )


def test_value_generation_ordinal():
    for i in (5, 12, 343, 6445, 22345, 344323):
        lst = generate_values(i)
        assert len(lst) == i
        assert all(isinstance(_, str) for _ in lst)
        assert all(int(_) == v for _, v in zip(lst, range(len(lst))))
        assert all(a > b for a, b in zip(lst[1:], lst[:-1]))


def test_can_create_different_types():
    assert isinstance(make_issue((0, 5)), ContiguousIssue)
    assert isinstance(make_issue(10), ContiguousIssue)
    assert isinstance(make_issue([1, 2, 3, 4, 5]), DiscreteCardinalIssue)
    assert isinstance(make_issue(["a", "b", "c"]), CategoricalIssue)
    assert isinstance(make_issue(lambda: 1), CallableIssue)
    assert isinstance(make_issue((0.0, 5.0)), ContinuousIssue)


def test_contains_discrete_list():
    issue = make_issue([0, 2, 10])
    assert 0 in issue
    assert 11 not in issue
    assert 0.0 in issue
    assert "abc" not in issue


def test_contains_discrete():
    for v in (10, (0, 9)):
        issue = make_issue(v)
        assert 0 in issue
        assert 11 not in issue
        assert 0.0 in issue
        assert "abc" not in issue


def test_contains_continuous():
    issue = make_issue((0, 5.0))
    assert 0 in issue
    assert 11 not in issue
    assert 4.56227344383 in issue
    assert "abc" not in issue


@pytest.mark.parametrize(
    "v", (10, (-1, 5), [1, 3, 4, 5], ["a", "b", "c", 5], [(0, 1), (8, 6)])
)
def test_can_loop_over_discrete_issue(v):
    issue = make_issue(v)
    assert issue.is_discrete()
    assert len(issue) > 0  # type: ignore I know that this is a discrete issue and len is applicable to it
    for _ in issue:
        pass


def test_can_loop_over_continuous_issue():
    issue = make_issue((0.0, 1.0))
    assert not issue.is_discrete()
    with pytest.raises(TypeError):
        assert len(issue) > 0  # type: ignore I know that this is a discrete issue and len is applicable to it
    for _ in issue:
        pass


def test_values_contained_in_issues_contiguous():
    i1, i2 = make_issue(10), make_issue(5)
    i3 = make_issue([1, 3, 4, 7])
    i4 = make_issue(["1", "2"])
    i5 = make_issue((0, 10))
    for v in i1:
        assert v in i1

    assert i2 in i1
    assert not i1 in i2
    assert i3 in i1
    assert not i1 in i3
    assert not i2 in i3
    for ix in (i1, i2, i3):
        assert not i4 in ix
        assert not ix in i4

    for ix in (i1, i2, i3):
        assert ix in i5
        assert i5 not in ix

    assert i4 not in i5
    assert i5 not in i4

    for v in i2:
        assert v in i1
    for v in i3:
        assert v in i1

    for x in (i1, i2, i3, i4, i5):
        assert x in x


def test_values_contained_in_outcome_spaces():
    a = make_os((make_issue(10), make_issue(list("abcdefg")), make_issue((1.0, 4.0))))
    b = make_os((make_issue(10), make_issue(list("adefg")), make_issue((1.0, 3.0))))
    c = make_os((make_issue(10), make_issue(list("adefgz")), make_issue((1.0, 3.0))))
    assert (4, "d", 3.5) in a
    assert (4, "d", 13.5) not in a

    for x in (a, b, c):
        assert x in x
    assert b in a
    assert not a in b
    assert c not in a

    assert not make_issue(10) in a
    assert make_issue(10, name=a.issues[0].name) in a
    for i in a.issues:
        assert i in a
        assert not i in b
        assert i not in c


def test_ordinal_issue_with_multiple_int_types():
    make_issue([0, 1, np.asarray([10])[0]])
    with pytest.raises(TypeError):
        make_issue([0, 1.0, np.asarray([10])[0]])
    with pytest.raises(TypeError):
        DiscreteOrdinalIssue([0, "1", np.asarray([10])[0]])


def test_ordinal_issue_with_multiple_float_types():
    make_issue([0.0, 1.0, np.asarray([10.0])[0]])
    with pytest.raises(TypeError):
        DiscreteOrdinalIssue([0.0, "1.0", np.asarray([10.0])[0]])
    with pytest.raises(TypeError):
        make_issue([0, 1.0, np.asarray([10.0])[0]])
